"""Vacuum Eve Plus (roidmi.vacuum.v60)"""


import json
import logging
import math
from datetime import timedelta
from enum import Enum

import click

from .click_common import EnumType, command
from .miot_device import DeviceStatus, MiotDevice, MiotMapping
from .vacuumcontainers import DNDStatus

_LOGGER = logging.getLogger(__name__)

_MAPPING: MiotMapping = {
    "battery_level": {"siid": 3, "piid": 1},
    "charging_state": {"siid": 3, "piid": 2},
    "error_code": {"siid": 2, "piid": 2},
    "state": {"siid": 2, "piid": 1},
    "filter_life_level": {"siid": 10, "piid": 1},
    "filter_left_minutes": {"siid": 10, "piid": 2},
    "main_brush_left_minutes": {"siid": 11, "piid": 1},
    "main_brush_life_level": {"siid": 11, "piid": 2},
    "side_brushes_left_minutes": {"siid": 12, "piid": 1},
    "side_brushes_life_level": {"siid": 12, "piid": 2},
    "sensor_dirty_time_left_minutes": {
        "siid": 15,
        "piid": 1,
    },  # named brush_left_time in the spec
    "sensor_dirty_remaning_level": {"siid": 15, "piid": 2},
    "sweep_mode": {"siid": 14, "piid": 1},
    "fanspeed_mode": {"siid": 2, "piid": 4},
    "sweep_type": {"siid": 2, "piid": 8},
    "path_mode": {"siid": 13, "piid": 8},
    "mop_present": {"siid": 8, "piid": 1},
    "work_station_freq": {"siid": 8, "piid": 2},  # Range: [0, 3, 1]
    "timing": {"siid": 8, "piid": 6},
    "clean_area": {"siid": 8, "piid": 7},  # uint32
    # "uid": {"siid": 8, "piid": 8},  # str - This UID is unknown
    "auto_boost": {"siid": 8, "piid": 9},
    "forbid_mode": {"siid": 8, "piid": 10},  # str
    "water_level": {"siid": 8, "piid": 11},
    "total_clean_time_sec": {"siid": 8, "piid": 13},
    "total_clean_areas": {"siid": 8, "piid": 14},
    "clean_counts": {"siid": 8, "piid": 18},
    "clean_time_sec": {"siid": 8, "piid": 19},
    "double_clean": {"siid": 8, "piid": 20},
    # "edge_sweep": {"siid": 8, "piid": 21}, # 2021-07-11: Roidmi Eve is not changing behavior when this bool is changed
    "led_switch": {"siid": 8, "piid": 22},
    "lidar_collision": {"siid": 8, "piid": 23},
    "station_key": {"siid": 8, "piid": 24},
    "station_led": {"siid": 8, "piid": 25},
    "current_audio": {"siid": 8, "piid": 26},
    # "progress": {"siid": 8, "piid": 28}, # 2021-07-11: this is part of the spec, but not implemented in Roidme Eve
    "station_type": {"siid": 8, "piid": 29},  # uint32
    # "voice_conf": {"siid": 8, "piid": 30}, # Always return file not exist !!!
    # "switch_status": {"siid": 2, "piid": 10}, # Enum with only one value: Open
    "volume": {"siid": 9, "piid": 1},
    "mute": {"siid": 9, "piid": 2},
    "start": {"siid": 2, "aiid": 1},
    "stop": {"siid": 2, "aiid": 2},
    "start_room_sweep": {"siid": 2, "aiid": 3},
    "start_sweep": {"siid": 14, "aiid": 1},
    "home": {"siid": 3, "aiid": 1},
    "identify": {"siid": 8, "aiid": 1},
    "start_station_dust_collection": {"siid": 8, "aiid": 6},
    "set_voice": {"siid": 8, "aiid": 12},
    "reset_filter_life": {"siid": 10, "aiid": 1},
    "reset_main_brush_life": {"siid": 11, "aiid": 1},
    "reset_side_brushes_life": {"siid": 12, "aiid": 1},
    "reset_sensor_dirty_life": {"siid": 15, "aiid": 1},
}


class ChargingState(Enum):
    Unknown = -1
    Charging = 1
    Discharging = 2
    NotChargeable = 4


class FanSpeed(Enum):
    Unknown = -1
    Silent = 1
    Basic = 2
    Strong = 3
    FullSpeed = 4
    Sweep = 0


class SweepType(Enum):
    Unknown = -1
    Sweep = 0
    Mop = 1
    MopAndSweep = 2


class PathMode(Enum):
    Unknown = -1
    Normal = 0
    YMopping = 1
    RepeatMopping = 2


class WaterLevel(Enum):
    Unknown = -1
    First = 1
    Second = 2
    Three = 3
    Fourth = 4
    Mop = 0


class SweepMode(Enum):
    Unknown = -1
    Total = 1
    Area = 2
    Curpoint = 3
    Point = 4
    Smart = 7
    AmartArea = 8
    DepthTotal = 9
    AlongWall = 10
    Idle = 0


error_codes = {
    0: "NoFaults",
    1: "LowBatteryFindCharger",
    2: "LowBatteryAndPoweroff",
    3: "WheelRap",
    4: "CollisionError",
    5: "TileDoTask",
    6: "LidarPointError",
    7: "FrontWallError",
    8: "PsdDirty",
    9: "MiddleBrushFatal",
    10: "SideBrush",
    11: "FanSpeedError",
    12: "LidarCover",
    13: "GarbageBoxFull",
    14: "GarbageBoxOut",
    15: "GarbageBoxFullOut",
    16: "PhysicalTrapped",
    17: "PickUpDoTask",
    18: "NoWaterBoxDoTask",
    19: "WaterBoxEmpty",
    20: "CleanCannotArrive",
    21: "StartFormForbid",
    22: "Drop",
    23: "KitWaterPump",
    24: "FindChargerFailed",
    25: "LowPowerClean",
}


class RoidmiState(Enum):
    Unknown = -1
    Dormant = 1
    Idle = 2
    Paused = 3
    Sweeping = 4
    GoCharging = 5
    Charging = 6
    Error = 7
    Rfctrl = 8
    Fullcharge = 9
    Shutdown = 10
    FindChargerPause = 11


class RoidmiVacuumStatus(DeviceStatus):
    """Container for status reports from the vacuum."""

    def __init__(self, data):
        """
        Response (MIoT format) of a Roidme Eve Plus (roidmi.vacuum.v60)
        [
            {'did': 'battery_level', 'siid': 3, 'piid': 1},
            {'did': 'charging_state', 'siid': 3, 'piid': 2},
            {'did': 'error_code', 'siid': 2, 'piid': 2},
            {'did': 'state', 'siid': 2, 'piid': 1},
            {'did': 'filter_life_level', 'siid': 10, 'piid': 1},
            {'did': 'filter_left_minutes', 'siid': 10, 'piid': 2},
            {'did': 'main_brush_left_minutes', 'siid': 11, 'piid': 1},
            {'did': 'main_brush_life_level', 'siid': 11, 'piid': 2},
            {'did': 'side_brushes_left_minutes', 'siid': 12, 'piid': 1},
            {'did': 'side_brushes_life_level', 'siid': 12, 'piid': 2},
            {'did': 'sensor_dirty_time_left_minutes', 'siid': 15, 'piid': 1},
            {'did': 'sensor_dirty_remaning_level', 'siid': 15, 'piid': 2},
            {'did': 'sweep_mode', 'siid': 14, 'piid': 1},
            {'did': 'fanspeed_mode', 'siid': 2, 'piid': 4},
            {'did': 'sweep_type', 'siid': 2, 'piid': 8}
            {'did': 'path_mode', 'siid': 13, 'piid': 8},
            {'did': 'mop_present', 'siid': 8, 'piid': 1},
            {'did': 'work_station_freq', 'siid': 8, 'piid': 2},
            {'did': 'timing', 'siid': 8, 'piid': 6},
            {'did': 'clean_area', 'siid': 8, 'piid': 7},
            {'did': 'auto_boost', 'siid': 8, 'piid': 9},
            {'did': 'forbid_mode', 'siid': 8, 'piid': 10},
            {'did': 'water_level', 'siid': 8, 'piid': 11},
            {'did': 'total_clean_time_sec', 'siid': 8, 'piid': 13},
            {'did': 'total_clean_areas', 'siid': 8, 'piid': 14},
            {'did': 'clean_counts', 'siid': 8, 'piid': 18},
            {'did': 'clean_time_sec', 'siid': 8, 'piid': 19},
            {'did': 'double_clean', 'siid': 8, 'piid': 20},
            {'did': 'led_switch', 'siid': 8, 'piid': 22}
            {'did': 'lidar_collision', 'siid': 8, 'piid': 23},
            {'did': 'station_key', 'siid': 8, 'piid': 24},
            {'did': 'station_led', 'siid': 8, 'piid': 25},
            {'did': 'current_audio', 'siid': 8, 'piid': 26},
            {'did': 'station_type', 'siid': 8, 'piid': 29},
            {'did': 'volume', 'siid': 9, 'piid': 1},
            {'did': 'mute', 'siid': 9, 'piid': 2}
        ]

        """
        self.data = data

    @property
    def battery(self) -> int:
        """Remaining battery in percentage."""
        return self.data["battery_level"]

    @property
    def error_code(self) -> int:
        """Error code as returned by the device."""
        return int(self.data["error_code"])

    @property
    def error(self) -> str:
        """Human readable error description, see also :func:`error_code`."""
        try:
            return error_codes[self.error_code]
        except KeyError:
            return "Definition missing for error %s" % self.error_code

    @property
    def charging_state(self) -> ChargingState:
        """Charging state (Charging/Discharging)"""
        try:
            return ChargingState(self.data["charging_state"])
        except ValueError:
            _LOGGER.error("Unknown ChargingStats (%s)", self.data["charging_state"])
            return ChargingState.Unknown

    @property
    def sweep_mode(self) -> SweepMode:
        """Sweep mode point/area/total etc."""
        try:
            return SweepMode(self.data["sweep_mode"])
        except ValueError:
            _LOGGER.error("Unknown SweepMode (%s)", self.data["sweep_mode"])
            return SweepMode.Unknown

    @property
    def fan_speed(self) -> FanSpeed:
        """Current fan speed."""
        try:
            return FanSpeed(self.data["fanspeed_mode"])
        except ValueError:
            _LOGGER.error("Unknown FanSpeed (%s)", self.data["fanspeed_mode"])
            return FanSpeed.Unknown

    @property
    def sweep_type(self) -> SweepType:
        """Current sweep type sweep/mop/sweep&mop."""
        try:
            return SweepType(self.data["sweep_type"])
        except ValueError:
            _LOGGER.error("Unknown SweepType (%s)", self.data["sweep_type"])
            return SweepType.Unknown

    @property
    def path_mode(self) -> PathMode:
        """Current path-mode:  normal/y-mopping etc."""
        try:
            return PathMode(self.data["path_mode"])
        except ValueError:
            _LOGGER.error("Unknown PathMode (%s)", self.data["path_mode"])
            return PathMode.Unknown

    @property
    def is_mop_attached(self) -> bool:
        """Return True if mop is attached."""
        return self.data["mop_present"]

    @property
    def dust_collection_frequency(self) -> int:
        """Frequency for emptying the dust bin.

        Example: 2 means the dust bin is emptied every second cleaning.
        """
        return self.data["work_station_freq"]

    @property
    def timing(self) -> str:
        """Repeated cleaning
        Example: {"time":[[32400,1,3,0,[1,2,3,4,5],0,[12,10],null],[57600,0,1,2,[1,2,3,4,5,6,0],2,[],null]],"tz":2,"tzs":7200}
        Cleaning 1:
            32400 = startTime(9:00)
            1=Enabled
            3=FanSpeed.Strong
            0=SweepType.Sweep
            [1,2,3,4,5]=Monday-Friday
            0=WaterLevel
            [12,10]=List of rooms
            null: ?Might be related to "Customize"?
        Cleaning 2:
            57600 = startTime(16:00)
            0=Disabled
            1=FanSpeed.Silent
            2=SweepType.MopAndSweep
            [1,2,3,4,5,6,0]=Monday-Sunday
            2=WaterLevel.Second
            []=All rooms
            null: ?Might be related to "Customize"?
        tz/tzs= time-zone
        """
        return self.data["timing"]

    @property
    def carpet_mode(self) -> bool:
        """Auto boost on carpet."""
        return self.data["auto_boost"]

    def _parse_forbid_mode(self, val) -> DNDStatus:
        # Example data: {"time":[75600,21600,1],"tz":2,"tzs":7200}
        def _seconds_to_components(val):
            hour = math.floor(val / 3600)
            minut = math.floor((val - hour * 3600) / 60)
            return (hour, minut)

        as_dict = json.loads(val)
        enabled = bool(as_dict["time"][2])
        start = _seconds_to_components(as_dict["time"][0])
        end = _seconds_to_components(as_dict["time"][1])
        return DNDStatus(
            dict(
                enabled=enabled,
                start_hour=start[0],
                start_minute=start[1],
                end_hour=end[0],
                end_minute=end[1],
            )
        )

    @property
    def dnd_status(self) -> DNDStatus:
        """Returns do-not-disturb status."""
        return self._parse_forbid_mode(self.data["forbid_mode"])

    @property
    def water_level(self) -> WaterLevel:
        """Get current water level."""
        try:
            return WaterLevel(self.data["water_level"])
        except ValueError:
            _LOGGER.error("Unknown WaterLevel (%s)", self.data["water_level"])
            return WaterLevel.Unknown

    @property
    def double_clean(self) -> bool:
        """Is double clean enabled."""
        return self.data["double_clean"]

    @property
    def led(self) -> bool:
        """Return True if led/display on vaccum is on."""
        return self.data["led_switch"]

    @property
    def is_lidar_collision_sensor(self) -> bool:
        """When ON, the robot will use lidar as the main detection sensor to help reduce
        collisions."""
        return self.data["lidar_collision"]

    @property
    def station_key(self) -> bool:
        """When ON: long press the display will turn on dust collection."""
        return self.data["station_key"]

    @property
    def station_led(self) -> bool:
        """Return if station display is on."""
        return self.data["station_led"]

    @property
    def current_audio(self) -> str:
        """Current voice setting.

        E.g. 'girl_en'
        """
        return self.data["current_audio"]

    @property
    def clean_time(self) -> timedelta:
        """Time used for cleaning (if finished, shows how long it took)."""
        return timedelta(seconds=self.data["clean_time_sec"])

    @property
    def clean_area(self) -> int:
        """Cleaned area in m2."""
        return self.data["clean_area"]

    @property
    def state_code(self) -> int:
        """State code as returned by the device."""
        return int(self.data["state"])

    @property
    def state(self) -> RoidmiState:
        """Human readable state description, see also :func:`state_code`."""
        try:
            return RoidmiState(self.state_code)
        except ValueError:
            _LOGGER.error("Unknown RoidmiState (%s)", self.state_code)
            return RoidmiState.Unknown

    @property
    def volume(self) -> int:
        """Return device sound volumen level."""
        return self.data["volume"]

    @property
    def is_muted(self) -> bool:
        """True if device is muted."""
        return bool(self.data["mute"])

    @property
    def is_paused(self) -> bool:
        """Return True if vacuum is paused."""
        return self.state in [RoidmiState.Paused, RoidmiState.FindChargerPause]

    @property
    def is_on(self) -> bool:
        """True if device is currently cleaning in any mode."""
        return self.state == RoidmiState.Sweeping

    @property
    def got_error(self) -> bool:
        """True if an error has occured."""
        return self.error_code != 0


class RoidmiCleaningSummary(DeviceStatus):
    """Contains summarized information about available cleaning runs."""

    def __init__(self, data) -> None:
        self.data = data

    @property
    def total_duration(self) -> timedelta:
        """Total cleaning duration."""
        return timedelta(seconds=self.data["total_clean_time_sec"])

    @property
    def total_area(self) -> int:
        """Total cleaned area."""
        return self.data["total_clean_areas"]

    @property
    def count(self) -> int:
        """Number of cleaning runs."""
        return self.data["clean_counts"]


class RoidmiConsumableStatus(DeviceStatus):
    """Container for consumable status information, including information about brushes
    and duration until they should be changed.

    The methods returning time left are based values returned from the device.
    """

    def __init__(self, data):
        self.data = data

    def _calcUsageTime(
        self, renaning_time: timedelta, remaning_level: int
    ) -> timedelta:
        remaning_fraction = remaning_level / 100.0
        original_total = renaning_time / remaning_fraction
        return original_total * (1 - remaning_fraction)

    @property
    def filter(self) -> timedelta:
        """Filter usage time."""
        return self._calcUsageTime(self.filter_left, self.data["filter_life_level"])

    @property
    def filter_left(self) -> timedelta:
        """How long until the filter should be changed."""
        return timedelta(minutes=self.data["filter_left_minutes"])

    @property
    def main_brush(self) -> timedelta:
        """Main brush usage time."""
        return self._calcUsageTime(
            self.main_brush_left, self.data["main_brush_life_level"]
        )

    @property
    def main_brush_left(self) -> timedelta:
        """How long until the main brush should be changed."""
        return timedelta(minutes=self.data["main_brush_left_minutes"])

    @property
    def side_brush(self) -> timedelta:
        """Main brush usage time."""
        return self._calcUsageTime(
            self.side_brush_left, self.data["side_brushes_life_level"]
        )

    @property
    def side_brush_left(self) -> timedelta:
        """How long until the side brushes should be changed."""
        return timedelta(minutes=self.data["side_brushes_left_minutes"])

    @property
    def sensor_dirty(self) -> timedelta:
        """Return time since last sensor clean."""
        return self._calcUsageTime(
            self.sensor_dirty_left, self.data["sensor_dirty_remaning_level"]
        )

    @property
    def sensor_dirty_left(self) -> timedelta:
        """How long until the sensors should be cleaned."""
        return timedelta(minutes=self.data["sensor_dirty_time_left_minutes"])


class RoidmiVacuumMiot(MiotDevice):
    """Interface for Vacuum Eve Plus (roidmi.vacuum.v60)"""

    mapping = _MAPPING

    @command()
    def status(self) -> RoidmiVacuumStatus:
        """State of the vacuum."""
        return RoidmiVacuumStatus(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                # max_properties limmit to 10 to avoid "Checksum error" messages from the device.
                for prop in self.get_properties_for_mapping()
            }
        )

    @command()
    def consumable_status(self) -> RoidmiConsumableStatus:
        """Return information about consumables."""
        return RoidmiConsumableStatus(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                # max_properties limmit to 10 to avoid "Checksum error" messages from the device.
                for prop in self.get_properties_for_mapping()
            }
        )

    @command()
    def cleaning_summary(self) -> RoidmiCleaningSummary:
        """Return information about cleaning runs."""
        return RoidmiCleaningSummary(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                # max_properties limmit to 10 to avoid "Checksum error" messages from the device.
                for prop in self.get_properties_for_mapping()
            }
        )

    @command()
    def start(self) -> None:
        """Start cleaning."""
        return self.call_action("start")

    # @command(click.argument("roomstr", type=str, required=False))
    # def start_room_sweep_unknown(self, roomstr: str=None) -> None:
    #     """Start room cleaning.

    #     roomstr: empty means start room clean of all rooms. FIXME: the syntax of an non-empty roomstr is still unknown
    #     """
    #     return self.call_action("start_room_sweep", roomstr)

    # @command(
    # click.argument("sweep_mode", type=EnumType(SweepMode)),
    # click.argument("clean_info", type=str),
    # )
    # def start_sweep_unknown(self, sweep_mode: SweepMode, clean_info: str=None) -> None:
    #     """Start sweep with mode.

    #     FIXME: the syntax of start_sweep is unknown
    #     """
    #     return self.call_action("start_sweep", [sweep_mode.value, clean_info])

    @command()
    def stop(self) -> None:
        """Stop cleaning."""
        return self.call_action("stop")

    @command()
    def home(self) -> None:
        """Return to home."""
        return self.call_action("home")

    @command()
    def identify(self) -> None:
        """Locate the device (i am here)."""
        return self.call_action("identify")

    @command(click.argument("on", type=bool))
    def set_station_led(self, on: bool):
        """Enable station led display."""
        return self.set_property("station_led", on)

    @command(click.argument("on", type=bool))
    def set_led(self, on: bool):
        """Enable vacuum led."""
        return self.set_property("led_switch", on)

    @command(click.argument("vol", type=int))
    def set_sound_volume(self, vol: int):
        """Set sound volume [0-100]."""
        return self.set_property("volume", vol)

    @command(click.argument("value", type=bool))
    def set_sound_muted(self, value: bool):
        """Set sound volume muted."""
        return self.set_property("mute", value)

    @command(click.argument("fanspeed_mode", type=EnumType(FanSpeed)))
    def set_fanspeed(self, fanspeed_mode: FanSpeed):
        """Set fan speed."""
        return self.set_property("fanspeed_mode", fanspeed_mode.value)

    @command(click.argument("sweep_type", type=EnumType(SweepType)))
    def set_sweep_type(self, sweep_type: SweepType):
        """Set sweep_type."""
        return self.set_property("sweep_type", sweep_type.value)

    @command(click.argument("path_mode", type=EnumType(PathMode)))
    def set_path_mode(self, path_mode: PathMode):
        """Set path_mode."""
        return self.set_property("path_mode", path_mode.value)

    @command(click.argument("dust_collection_frequency", type=int))
    def set_dust_collection_frequency(self, dust_collection_frequency: int):
        """Set frequency for emptying the dust bin.

        Example: 2 means the dust bin is emptied every second cleaning.
        """
        return self.set_property("work_station_freq", dust_collection_frequency)

    @command(click.argument("timing", type=str))
    def set_timing(self, timing: str):
        """Set repeated clean timing.

        Set timing to 9:00 Monday-Friday, rooms:[12,10]
        timing = '{"time":[[32400,1,3,0,[1,2,3,4,5],0,[12,10],null]],"tz":2,"tzs":7200}'
        See also :func:`RoidmiVacuumStatus.timing`

        NOTE: setting timing will override existing settings
        """
        return self.set_property("timing", timing)

    @command(click.argument("auto_boost", type=bool))
    def set_carpet_mode(self, auto_boost: bool):
        """Set auto boost on carpet."""
        return self.set_property("auto_boost", auto_boost)

    def _set_dnd(self, start_int: int, end_int: int, active: bool):
        value_str = json.dumps({"time": [start_int, end_int, int(active)]})
        return self.set_property("forbid_mode", value_str)

    @command(
        click.argument("start_hr", type=int),
        click.argument("start_min", type=int),
        click.argument("end_hr", type=int),
        click.argument("end_min", type=int),
    )
    def set_dnd(self, start_hr: int, start_min: int, end_hr: int, end_min: int):
        """Set do-not-disturb.

        :param int start_hr: Start hour
        :param int start_min: Start minute
        :param int end_hr: End hour
        :param int end_min: End minute
        """
        start_int = int(timedelta(hours=start_hr, minutes=start_min).total_seconds())
        end_int = int(timedelta(hours=end_hr, minutes=end_min).total_seconds())
        return self._set_dnd(start_int, end_int, active=True)

    @command()
    def disable_dnd(self):
        """Disable do-not-disturb."""
        # The current do not disturb is read back for a better user expierence,
        # as start/end time must be set together with enabled=False
        try:
            current_dnd_str = self.get_property_by(**_MAPPING["forbid_mode"])[0][
                "value"
            ]
            current_dnd_dict = json.loads(current_dnd_str)
        except Exception:
            # In case reading current DND back fails, DND is disabled anyway
            return self._set_dnd(0, 0, active=False)
        return self._set_dnd(
            current_dnd_dict["time"][0], current_dnd_dict["time"][1], active=False
        )

    @command(click.argument("water_level", type=EnumType(WaterLevel)))
    def set_water_level(self, water_level: WaterLevel):
        """Set water_level."""
        return self.set_property("water_level", water_level.value)

    @command(click.argument("double_clean", type=bool))
    def set_double_clean(self, double_clean: bool):
        """Set double clean (True/False)."""
        return self.set_property("double_clean", double_clean)

    @command(click.argument("lidar_collision", type=bool))
    def set_lidar_collision_sensor(self, lidar_collision: bool):
        """When ON, the robot will use lidar as the main detection sensor to help reduce
        collisions."""
        return self.set_property("lidar_collision", lidar_collision)

    @command()
    def start_dust(self) -> None:
        """Start base dust collection."""
        return self.call_action("start_station_dust_collection")

    # @command(click.argument("voice", type=str))
    #     def set_voice_unknown(self, voice: str) -> None:
    #     """Set voice.

    #     FIXME: the syntax of voice is unknown (assumed to be json format)
    #     """
    #     return self.call_action("set_voice", voice)

    @command()
    def reset_filter_life(self) -> None:
        """Reset filter life."""
        return self.call_action("reset_filter_life")

    @command()
    def reset_mainbrush_life(self) -> None:
        """Reset main brush life."""
        return self.call_action("reset_main_brush_life")

    @command()
    def reset_sidebrush_life(self) -> None:
        """Reset side brushes life."""
        return self.call_action("reset_side_brushes_life")

    @command()
    def reset_sensor_dirty_life(self) -> None:
        """Reset sensor dirty life."""
        return self.call_action("reset_sensor_dirty_life")



import ast
import contextlib
import json
import logging
import pathlib
import sys
import threading
import time
from pprint import pformat as pf
from typing import Any, List  # noqa: F401

import click
from appdirs import user_cache_dir
from tqdm import tqdm

import miio  # noqa: E402
from miio.click_common import (
    ExceptionHandlerGroup,
    LiteralParamType,
    validate_ip,
    validate_token,
)
from miio.device import UpdateState
from miio.exceptions import DeviceInfoUnavailableException
from miio.miioprotocol import MiIOProtocol
from miio.updater import OneShotServer
from miio.vacuum import CarpetCleaningMode

_LOGGER = logging.getLogger(__name__)
pass_dev = click.make_pass_decorator(miio.Device, ensure=True)


@click.group(invoke_without_command=True, cls=ExceptionHandlerGroup)
@click.option("--ip", envvar="MIROBO_IP", callback=validate_ip)
@click.option("--token", envvar="MIROBO_TOKEN", callback=validate_token)
@click.option("-d", "--debug", default=False, count=True)
@click.option(
    "--id-file",
    type=click.Path(dir_okay=False, writable=True),
    default=user_cache_dir("python-miio") + "/python-mirobo.seq",
)
@click.version_option()
@click.pass_context
def cli(ctx, ip: str, token: str, debug: int, id_file: str):
    """A tool to command Xiaomi Vacuum robot."""
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        _LOGGER.info("Debug mode active")
    else:
        logging.basicConfig(level=logging.INFO)

    # if we are scanning, we do not try to connect.
    if ctx.invoked_subcommand == "discover":
        ctx.obj = "discover"
        return

    if ip is None or token is None:
        click.echo("You have to give ip and token!")
        sys.exit(-1)

    start_id = manual_seq = 0
    with contextlib.suppress(FileNotFoundError, TypeError, ValueError), open(
        id_file, "r"
    ) as f:
        x = json.load(f)
        start_id = x.get("seq", 0)
        manual_seq = x.get("manual_seq", 0)
        _LOGGER.debug("Read stored sequence ids: %s", x)

    vac = miio.Vacuum(ip, token, start_id, debug)

    vac.manual_seqnum = manual_seq
    _LOGGER.debug("Connecting to %s with token %s", ip, token)

    ctx.obj = vac

    if ctx.invoked_subcommand is None:
        ctx.invoke(status)
        cleanup(vac, id_file=id_file)


@cli.resultcallback()
@pass_dev
def cleanup(vac: miio.Vacuum, *args, **kwargs):
    if vac.ip is None:  # dummy Device for discovery, skip teardown
        return
    id_file = kwargs["id_file"]
    seqs = {"seq": vac.raw_id, "manual_seq": vac.manual_seqnum}
    _LOGGER.debug("Writing %s to %s", seqs, id_file)

    path_obj = pathlib.Path(id_file)
    dir = path_obj.parents[0]
    dir.mkdir(parents=True, exist_ok=True)

    with open(id_file, "w") as f:
        json.dump(seqs, f)


@cli.command()
@click.option("--handshake", type=bool, default=False)
def discover(handshake):
    """Search for robots in the network."""
    if handshake:
        MiIOProtocol.discover()
    else:
        miio.Discovery.discover_mdns()


@cli.command()
@pass_dev
def status(vac: miio.Vacuum):
    """Returns the state information."""
    res = vac.status()
    if not res:
        return  # bail out

    if res.error_code:
        click.echo(click.style("Error: %s !" % res.error, bold=True, fg="red"))
    if res.is_water_shortage:
        click.echo(click.style("Water is running low!", bold=True, fg="blue"))
    click.echo(click.style("State: %s" % res.state, bold=True))
    click.echo("Battery: %s %%" % res.battery)
    click.echo("Fanspeed: %s %%" % res.fanspeed)
    click.echo("Cleaning since: %s" % res.clean_time)
    click.echo("Cleaned area: %s m²" % res.clean_area)
    # click.echo("DND enabled: %s" % res.dnd)
    # click.echo("Map present: %s" % res.map)
    # click.echo("in_cleaning: %s" % res.in_cleaning)
    click.echo("Water box attached: %s" % res.is_water_box_attached)
    if res.is_water_box_carriage_attached is not None:
        click.echo("Mop attached: %s" % res.is_water_box_carriage_attached)


@cli.command()
@pass_dev
def consumables(vac: miio.Vacuum):
    """Return consumables status."""
    res = vac.consumable_status()
    click.echo("Main brush:   %s (left %s)" % (res.main_brush, res.main_brush_left))
    click.echo("Side brush:   %s (left %s)" % (res.side_brush, res.side_brush_left))
    click.echo("Filter:       %s (left %s)" % (res.filter, res.filter_left))
    click.echo("Sensor dirty: %s (left %s)" % (res.sensor_dirty, res.sensor_dirty_left))


@cli.command()
@click.argument("name", type=str, required=True)
@pass_dev
def reset_consumable(vac: miio.Vacuum, name):
    """Reset consumable state.

    Allowed values: main_brush, side_brush, filter, sensor_dirty
    """
    from miio.vacuum import Consumable

    if name == "main_brush":
        consumable = Consumable.MainBrush
    elif name == "side_brush":
        consumable = Consumable.SideBrush
    elif name == "filter":
        consumable = Consumable.Filter
    elif name == "sensor_dirty":
        consumable = Consumable.SensorDirty
    else:
        click.echo("Unexpected state name: %s" % name)
        return

    click.echo(
        "Resetting consumable '%s': %s" % (name, vac.consumable_reset(consumable))
    )


@cli.command()
@pass_dev
def start(vac: miio.Vacuum):
    """Start cleaning."""
    click.echo("Starting cleaning: %s" % vac.start())


@cli.command()
@pass_dev
def spot(vac: miio.Vacuum):
    """Start spot cleaning."""
    click.echo("Starting spot cleaning: %s" % vac.spot())


@cli.command()
@pass_dev
def pause(vac: miio.Vacuum):
    """Pause cleaning."""
    click.echo("Pausing: %s" % vac.pause())


@cli.command()
@pass_dev
def stop(vac: miio.Vacuum):
    """Stop cleaning."""
    click.echo("Stop cleaning: %s" % vac.stop())


@cli.command()
@pass_dev
def home(vac: miio.Vacuum):
    """Return home."""
    click.echo("Requesting return to home: %s" % vac.home())


@cli.command()
@pass_dev
@click.argument("x_coord", type=int)
@click.argument("y_coord", type=int)
def goto(vac: miio.Vacuum, x_coord: int, y_coord: int):
    """Go to specific target."""
    click.echo("Going to target : %s" % vac.goto(x_coord, y_coord))


@cli.command()
@pass_dev
@click.argument("zones", type=LiteralParamType(), required=True)
def zoned_clean(vac: miio.Vacuum, zones: List):
    """Clean zone."""
    click.echo("Cleaning zone(s) : %s" % vac.zoned_clean(zones))


@cli.group()
@pass_dev
# @click.argument('command', required=False)
def manual(vac: miio.Vacuum):
    """Control the robot manually."""
    command = ""
    if command == "start":
        click.echo("Starting manual control")
        return vac.manual_start()
    if command == "stop":
        click.echo("Stopping manual control")
        return vac.manual_stop()
    # if not vac.manual_mode and command :


@manual.command()
@pass_dev
def tui(vac: miio.Vacuum):
    """TUI for the manual mode."""
    miio.VacuumTUI(vac).run()


@manual.command(name="start")
@pass_dev
def manual_start(vac: miio.Vacuum):  # noqa: F811  # redef of start
    """Activate the manual mode."""
    click.echo("Activating manual controls")
    return vac.manual_start()


@manual.command(name="stop")
@pass_dev
def manual_stop(vac: miio.Vacuum):  # noqa: F811  # redef of stop
    """Deactivate the manual mode."""
    click.echo("Deactivating manual controls")
    return vac.manual_stop()


@manual.command()
@pass_dev
@click.argument("degrees", type=int)
def left(vac: miio.Vacuum, degrees: int):
    """Turn to left."""
    click.echo("Turning %s degrees left" % degrees)
    return vac.manual_control(degrees, 0)


@manual.command()
@pass_dev
@click.argument("degrees", type=int)
def right(vac: miio.Vacuum, degrees: int):
    """Turn to right."""
    click.echo("Turning right")
    return vac.manual_control(-degrees, 0)


@manual.command()
@click.argument("amount", type=float)
@pass_dev
def forward(vac: miio.Vacuum, amount: float):
    """Run forwards."""
    click.echo("Moving forwards")
    return vac.manual_control(0, amount)


@manual.command()
@click.argument("amount", type=float)
@pass_dev
def backward(vac: miio.Vacuum, amount: float):
    """Run backwards."""
    click.echo("Moving backwards")
    return vac.manual_control(0, -amount)


@manual.command()
@pass_dev
@click.argument("rotation", type=float)
@click.argument("velocity", type=float)
@click.argument("duration", type=int)
def move(vac: miio.Vacuum, rotation: int, velocity: float, duration: int):
    """Pass raw manual values."""
    return vac.manual_control(rotation, velocity, duration)


@cli.command()
@click.argument("cmd", required=False)
@click.argument("start_hr", type=int, required=False)
@click.argument("start_min", type=int, required=False)
@click.argument("end_hr", type=int, required=False)
@click.argument("end_min", type=int, required=False)
@pass_dev
def dnd(
    vac: miio.Vacuum, cmd: str, start_hr: int, start_min: int, end_hr: int, end_min: int
):
    """Query and adjust do-not-disturb mode."""
    if cmd == "off":
        click.echo("Disabling DND..")
        click.echo(vac.disable_dnd())
    elif cmd == "on":
        click.echo(
            "Enabling DND %s:%s to %s:%s" % (start_hr, start_min, end_hr, end_min)
        )
        click.echo(vac.set_dnd(start_hr, start_min, end_hr, end_min))
    else:
        x = vac.dnd_status()
        click.echo(
            click.style(
                "Between %s and %s (enabled: %s)" % (x.start, x.end, x.enabled),
                bold=x.enabled,
            )
        )


@cli.command()
@click.argument("speed", type=int, required=False)
@pass_dev
def fanspeed(vac: miio.Vacuum, speed):
    """Query and adjust the fan speed."""
    if speed:
        click.echo("Setting fan speed to %s" % speed)
        vac.set_fan_speed(speed)
    else:
        click.echo("Current fan speed: %s" % vac.fan_speed())


@cli.group(invoke_without_command=True)
@pass_dev
@click.pass_context
def timer(ctx, vac: miio.Vacuum):
    """List and modify existing timers."""
    if ctx.invoked_subcommand is not None:
        return
    timers = vac.timer()
    click.echo("Timezone: %s\n" % vac.timezone())
    for idx, timer in enumerate(timers):
        color = "green" if timer.enabled else "yellow"
        click.echo(
            click.style(
                "Timer #%s, id %s (ts: %s)" % (idx, timer.id, timer.ts),
                bold=True,
                fg=color,
            )
        )
        click.echo("  %s" % timer.cron)
        min, hr, x, y, days = timer.cron.split(" ")
        cron = "%s %s %s %s %s" % (min, hr, x, y, days)
        click.echo("  %s" % cron)


@timer.command()
@click.option("--cron")
@click.option("--command", default="", required=False)
@click.option("--params", default="", required=False)
@pass_dev
def add(vac: miio.Vacuum, cron, command, params):
    """Add a timer."""
    click.echo(vac.add_timer(cron, command, params))


@timer.command()
@click.argument("timer_id", type=int, required=True)
@pass_dev
def delete(vac: miio.Vacuum, timer_id):
    """Delete a timer."""
    click.echo(vac.delete_timer(timer_id))


@timer.command()
@click.argument("timer_id", type=int, required=True)
@click.option("--enable", is_flag=True)
@click.option("--disable", is_flag=True)
@pass_dev
def update(vac: miio.Vacuum, timer_id, enable, disable):
    """Enable/disable a timer."""
    from miio.vacuum import TimerState

    if enable and not disable:
        vac.update_timer(timer_id, TimerState.On)
    elif disable and not enable:
        vac.update_timer(timer_id, TimerState.Off)
    else:
        click.echo("You need to specify either --enable or --disable")


@cli.command()
@pass_dev
def find(vac: miio.Vacuum):
    """Find the robot."""
    click.echo("Sending find the robot calls.")
    click.echo(vac.find())


@cli.command()
@pass_dev
def map(vac: miio.Vacuum):
    """Return the map token."""
    click.echo(vac.map())


@cli.command()
@pass_dev
def info(vac: miio.Vacuum):
    """Return device information."""
    try:
        res = vac.info()

        click.echo("%s" % res)
        _LOGGER.debug("Full response: %s", pf(res.raw))
    except DeviceInfoUnavailableException:
        click.echo(
            "Unable to fetch info, this can happen when the vacuum "
            "is not connected to the Xiaomi cloud."
        )


@cli.command()
@pass_dev
def cleaning_history(vac: miio.Vacuum):
    """Query the cleaning history."""
    res = vac.clean_history()
    click.echo("Total clean count: %s" % res.count)
    click.echo("Cleaned for: %s (area: %s m²)" % (res.total_duration, res.total_area))
    if res.dust_collection_count is not None:
        click.echo("Emptied dust collection bin: %s times" % res.dust_collection_count)
    click.echo()
    for idx, id_ in enumerate(res.ids):
        details = vac.clean_details(id_, return_list=False)
        color = "green" if details.complete else "yellow"
        click.echo(
            click.style(
                "Clean #%s: %s-%s (complete: %s, error: %s)"
                % (idx, details.start, details.end, details.complete, details.error),
                bold=True,
                fg=color,
            )
        )
        click.echo("  Area cleaned: %s m²" % details.area)
        click.echo("  Duration: (%s)" % details.duration)
        click.echo()


@cli.command()
@click.argument("volume", type=int, required=False)
@click.option("--test", "test_mode", is_flag=True, help="play a test tune")
@pass_dev
def sound(vac: miio.Vacuum, volume: int, test_mode: bool):
    """Query and change sound settings."""
    if volume is not None:
        click.echo("Setting sound volume to %s" % volume)
        vac.set_sound_volume(volume)
    if test_mode:
        vac.test_sound_volume()
    click.echo("Current sound: %s" % vac.sound_info())
    click.echo("Current volume: %s" % vac.sound_volume())
    click.echo("Install progress: %s" % vac.sound_install_progress())


@cli.command()
@click.argument("url")
@click.argument("md5sum", required=False, default=None)
@click.option("--sid", type=int, required=False, default=10000)
@click.option("--ip", required=False)
@pass_dev
def install_sound(vac: miio.Vacuum, url: str, md5sum: str, sid: int, ip: str):
    """Install a sound.

    When passing a local file this will create a self-hosting server
     for the given file and the md5sum will be calculated automatically.

    For URLs you have to specify the md5sum manually.

    `--ip` can be used to override automatically detected IP address for
     the device to contact for the update.
    """
    click.echo("Installing from %s (md5: %s) for id %s" % (url, md5sum, sid))

    local_url = None
    server = None
    if url.startswith("http"):
        if md5sum is None:
            click.echo("You need to pass md5 when using URL for updating.")
            return
        local_url = url
    else:
        server = OneShotServer(url)
        local_url = server.url(ip)
        md5sum = server.md5

        t = threading.Thread(target=server.serve_once)
        t.start()
        click.echo("Hosting file at %s" % local_url)

    click.echo(vac.install_sound(local_url, md5sum, sid))

    progress = vac.sound_install_progress()
    while progress.is_installing:
        progress = vac.sound_install_progress()
        click.echo("%s (%s %%)" % (progress.state.name, progress.progress))
        time.sleep(1)

    progress = vac.sound_install_progress()

    if progress.is_errored:
        click.echo("Error during installation: %s" % progress.error)
    else:
        click.echo("Installation of sid '%s' complete!" % sid)

    if server is not None:
        t.join()


@cli.command()
@pass_dev
def serial_number(vac: miio.Vacuum):
    """Query serial number."""
    click.echo("Serial#: %s" % vac.serial_number())


@cli.command()
@click.argument("tz", required=False)
@pass_dev
def timezone(vac: miio.Vacuum, tz=None):
    """Query or set the timezone."""
    if tz is not None:
        click.echo("Setting timezone to: %s" % tz)
        click.echo(vac.set_timezone(tz))
    else:
        click.echo("Timezone: %s" % vac.timezone())


@cli.command()
@click.argument("enabled", required=False, type=bool)
@pass_dev
def carpet_mode(vac: miio.Vacuum, enabled=None):
    """Query or set the carpet mode."""
    if enabled is None:
        click.echo(vac.carpet_mode())
    else:
        click.echo(vac.set_carpet_mode(enabled))


@cli.command()
@click.argument("mode", required=False, type=str)
@pass_dev
def carpet_cleaning_mode(vac: miio.Vacuum, mode=None):
    """Query or set the carpet cleaning/avoidance mode.

    Allowed values: Avoid, Rise, Ignore
    """

    if mode is None:
        click.echo("Carpet cleaning mode: %s" % vac.carpet_cleaning_mode())
    else:
        click.echo(
            "Setting carpet cleaning mode: %s"
            % vac.set_carpet_cleaning_mode(CarpetCleaningMode[mode])
        )


@cli.command()
@click.argument("ssid", required=True)
@click.argument("password", required=True)
@click.argument("uid", type=int, required=False)
@click.option("--timezone", type=str, required=False, default=None)
@pass_dev
def configure_wifi(vac: miio.Vacuum, ssid: str, password: str, uid: int, timezone: str):
    """Configure the wifi settings.

    Note that some newer firmwares may expect you to define the timezone by using
    --timezone.
    """
    click.echo("Configuring wifi to SSID: %s" % ssid)
    click.echo(vac.configure_wifi(ssid, password, uid, timezone))


@cli.command()
@pass_dev
def update_status(vac: miio.Vacuum):
    """Return update state and progress."""
    update_state = vac.update_state()
    click.echo("Update state: %s" % update_state)

    if update_state == UpdateState.Downloading:
        click.echo("Update progress: %s" % vac.update_progress())


@cli.command()
@click.argument("url", required=True)
@click.argument("md5", required=False, default=None)
@click.option("--ip", required=False)
@pass_dev
def update_firmware(vac: miio.Vacuum, url: str, md5: str, ip: str):
    """Update device firmware.

    If `url` starts with http* it is expected to be an URL.
     In that case md5sum of the file has to be given.

    `--ip` can be used to override automatically detected IP address for
     the device to contact for the update.
    """

    # TODO Check that the device is in updateable state.

    click.echo("Going to update from %s" % url)
    if url.lower().startswith("http"):
        if md5 is None:
            click.echo("You need to pass md5 when using URL for updating.")
            return

        click.echo("Using %s (md5: %s)" % (url, md5))
    else:
        server = OneShotServer(url)
        url = server.url(ip)

        t = threading.Thread(target=server.serve_once)
        t.start()
        click.echo("Hosting file at %s" % url)
        md5 = server.md5

    update_res = vac.update(url, md5)
    if update_res:
        click.echo("Update started!")
    else:
        click.echo("Starting the update failed: %s" % update_res)

    with tqdm(total=100) as pbar:
        state = vac.update_state()
        while state == UpdateState.Downloading:
            try:
                state = vac.update_state()
                progress = vac.update_progress()
            except:  # noqa # nosec
                # we may not get our messages through during uploads
                continue

            if state == UpdateState.Installing:
                click.echo("Installation started, please wait until the vacuum reboots")
                break

            pbar.update(progress - pbar.n)
            pbar.set_description("%s" % state.name)
            time.sleep(1)


@cli.command()
@click.argument("cmd", required=True)
@click.argument("parameters", required=False)
@pass_dev
def raw_command(vac: miio.Vacuum, cmd, parameters):
    """Run a raw command."""
    params = []  # type: Any
    if parameters:
        params = ast.literal_eval(parameters)
    click.echo("Sending cmd %s with params %s" % (cmd, params))
    click.echo(vac.raw_command(cmd, params))


if __name__ == "__main__":
    cli()



class DeviceException(Exception):
    """Exception wrapping any communication errors with the device."""


class PayloadDecodeException(DeviceException):
    """Exception for failures in payload decoding.

    This is raised when the json payload cannot be decoded, indicating invalid response
    from a device.
    """


class DeviceInfoUnavailableException(DeviceException):
    """Exception raised when requesting miio.info fails.

    This allows users to gracefully handle cases where the information unavailable. This
    can happen, for instance, when the device has no cloud access.
    """


class DeviceError(DeviceException):
    """Exception communicating an error delivered by the target device.

    The device given error code and message can be accessed with  `code` and `message`
    variables.
    """

    def __init__(self, error):
        self.code = error.get("code")
        self.message = error.get("message")


class RecoverableError(DeviceError):
    """Exception communicating an recoverable error delivered by the target device."""



import enum
import logging
from datetime import timedelta
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)


class WalkingpadException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Auto = 0
    Manual = 1
    Off = 2


class OperationSensitivity(enum.Enum):
    High = 1
    Medium = 2
    Low = 3


class WalkingpadStatus(DeviceStatus):
    """Container for status reports from Xiaomi Walkingpad A1 (ksmb.walkingpad.v3).

    Input data dictionary to initialise this class:

    {'cal': 6130,
     'dist': 90,
     'mode': 1,
     'power': 'on',
     'sensitivity': 1,
     'sp': 3.0,
     'start_speed': 3.0,
     'step': 180,
     'time': 121}
    """

    def __init__(self, data: Dict[str, Any]) -> None:
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.power == "on"

    @property
    def walking_time(self) -> timedelta:
        """Current walking duration in seconds."""
        return timedelta(seconds=int(self.data["time"]))

    @property
    def speed(self) -> float:
        """Current speed."""
        return float(self.data["sp"])

    @property
    def start_speed(self) -> float:
        """Current start speed."""
        return self.data["start_speed"]

    @property
    def mode(self) -> OperationMode:
        """Current mode."""
        return OperationMode(self.data["mode"])

    @property
    def sensitivity(self) -> OperationSensitivity:
        """Current sensitivity."""
        return OperationSensitivity(self.data["sensitivity"])

    @property
    def step_count(self) -> int:
        """Current steps."""
        return int(self.data["step"])

    @property
    def distance(self) -> int:
        """Current distance in meters."""
        return int(self.data["dist"])

    @property
    def calories(self) -> int:
        """Current calories burnt."""
        return int(self.data["cal"])


class Walkingpad(Device):
    """Main class representing Xiaomi Walkingpad."""

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode.name}\n"
            "Time: {result.walking_time}\n"
            "Steps: {result.step_count}\n"
            "Speed: {result.speed}\n"
            "Start Speed: {result.start_speed}\n"
            "Sensitivity: {result.sensitivity.name}\n"
            "Distance: {result.distance}\n"
            "Calories: {result.calories}",
        )
    )
    def status(self) -> WalkingpadStatus:
        """Retrieve properties."""

        data = self._get_quick_status()

        # The quick status only retrieves a subset of the properties. The rest of them are retrieved here.
        properties_additional = ["power", "mode", "start_speed", "sensitivity"]
        values_additional = self.get_properties(properties_additional, max_properties=1)

        additional_props = dict(zip(properties_additional, values_additional))
        data.update(additional_props)

        return WalkingpadStatus(data)

    @command(
        default_output=format_output(
            "",
            "Mode: {result.mode.name}\n"
            "Walking time: {result.walking_time}\n"
            "Steps: {result.step_count}\n"
            "Speed: {result.speed}\n"
            "Distance: {result.distance}\n"
            "Calories: {result.calories}",
        )
    )
    def quick_status(self) -> WalkingpadStatus:
        """Retrieve quick status.

        The walkingpad provides the option to retrieve a subset of properties in one call:
        steps, mode, speed, distance, calories and time.

        `status()` will do four more separate I/O requests for power, mode, start_speed, and sensitivity.
        If you don't need any of that, prefer this method for status updates.
        """

        data = self._get_quick_status()

        return WalkingpadStatus(data)

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(default_output=format_output("Locking"))
    def lock(self):
        """Lock device."""
        return self.send("set_lock", [1])

    @command(default_output=format_output("Unlocking"))
    def unlock(self):
        """Unlock device."""
        return self.send("set_lock", [0])

    @command(default_output=format_output("Starting the treadmill"))
    def start(self):
        """Start the treadmill."""

        # In case the treadmill is not already turned on, turn it on.
        if not self.status().is_on:
            self.on()

        return self.send("set_state", ["run"])

    @command(default_output=format_output("Stopping the treadmill"))
    def stop(self):
        """Stop the treadmill."""
        return self.send("set_state", ["stop"])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.name}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode (auto/manual)."""

        if not isinstance(mode, OperationMode):
            raise WalkingpadException("Invalid mode: %s" % mode)

        return self.send("set_mode", [mode.value])

    @command(
        click.argument("speed", type=float),
        default_output=format_output("Setting speed to {speed}"),
    )
    def set_speed(self, speed: float):
        """Set speed."""

        # In case the treadmill is not already turned on, throw an exception.
        if not self.status().is_on:
            raise WalkingpadException("Cannot set the speed, device is turned off")

        if not isinstance(speed, float):
            raise WalkingpadException("Invalid speed: %s" % speed)

        if speed < 0 or speed > 6:
            raise WalkingpadException("Invalid speed: %s" % speed)

        return self.send("set_speed", [speed])

    @command(
        click.argument("speed", type=float),
        default_output=format_output("Setting start speed to {speed}"),
    )
    def set_start_speed(self, speed: float):
        """Set start speed."""

        # In case the treadmill is not already turned on, throw an exception.
        if not self.status().is_on:
            raise WalkingpadException(
                "Cannot set the start speed, device is turned off"
            )

        if not isinstance(speed, float):
            raise WalkingpadException("Invalid start speed: %s" % speed)

        if speed < 0 or speed > 6:
            raise WalkingpadException("Invalid start speed: %s" % speed)

        return self.send("set_start_speed", [speed])

    @command(
        click.argument("sensitivity", type=EnumType(OperationSensitivity)),
        default_output=format_output("Setting sensitivity to {sensitivity}"),
    )
    def set_sensitivity(self, sensitivity: OperationSensitivity):
        """Set sensitivity."""

        if not isinstance(sensitivity, OperationSensitivity):
            raise WalkingpadException("Invalid mode: %s" % sensitivity)

        return self.send("set_sensitivity", [sensitivity.value])

    def _get_quick_status(self):
        """Internal helper to get the quick status via the "all" property."""

        # Walkingpad A1 allows you to quickly retrieve a subset of values with "all"
        # all other properties need to be retrieved one by one and are therefore slower
        # eg ['mode:1', 'time:1387', 'sp:3.0', 'dist:1150', 'cal:71710', 'step:2117']

        properties = ["all"]

        values = self.get_properties(properties, max_properties=1)

        value_map = {
            "sp": float,
            "step": int,
            "cal": int,
            "time": int,
            "dist": int,
            "mode": int,
        }

        data = {}
        for x in values:
            prop, value = x.split(":")

            if prop not in value_map:
                _LOGGER.warning("Received unknown data from device: %s=%s", prop, value)

            data[prop] = value

        converted_data = {key: value_map[key](value) for key, value in data.items()}

        return converted_data



from unittest import TestCase

import pytest

from miio import AirDehumidifier
from miio.airdehumidifier import (
    MODEL_DEHUMIDIFIER_V1,
    AirDehumidifierException,
    AirDehumidifierStatus,
    FanSpeed,
    OperationMode,
)
from miio.device import DeviceInfo

from .dummies import DummyDevice


class DummyAirDehumidifierV1(DummyDevice, AirDehumidifier):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_DEHUMIDIFIER_V1
        self.dummy_device_info = {
            "life": 348202,
            "uid": 1759530000,
            "model": "nwt.derh.wdh318efw1",
            "token": "68ffffffffffffffffffffffffffffff",
            "fw_ver": "2.0.5",
            "mcu_fw_ver": "0018",
            "miio_ver": "0.0.5",
            "hw_ver": "esp32",
            "mmfree": 65476,
            "mac": "78:11:FF:FF:FF:FF",
            "wifi_fw_ver": "v3.1.4-56-g8ffb04960",
            "netif": {
                "gw": "192.168.0.1",
                "localIp": "192.168.0.25",
                "mask": "255.255.255.0",
            },
        }

        self.device_info = None

        self.state = {
            "on_off": "on",
            "mode": "auto",
            "fan_st": 2,
            "buzzer": "off",
            "led": "on",
            "child_lock": "off",
            "humidity": 48,
            "temp": 34,
            "compressor_status": "off",
            "fan_speed": 0,
            "tank_full": "off",
            "defrost_status": "off",
            "alarm": "ok",
            "auto": 50,
        }

        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("on_off", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_led": lambda x: self._set_state("led", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_fan_speed": lambda x: self._set_state("fan_st", x),
            "set_auto": lambda x: self._set_state("auto", x),
            "miIO.info": self._get_device_info,
        }
        super().__init__(args, kwargs)

    def _get_device_info(self, _):
        """Return dummy device info."""
        return self.dummy_device_info


@pytest.fixture(scope="class")
def airdehumidifierv1(request):
    request.cls.device = DummyAirDehumidifierV1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airdehumidifierv1")
class TestAirDehumidifierV1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        device_info = DeviceInfo(self.device.dummy_device_info)

        assert repr(self.state()) == repr(
            AirDehumidifierStatus(self.device.start_state, device_info)
        )

        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["temp"]
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().led == (self.device.start_state["led"] == "on")
        assert self.state().buzzer == (self.device.start_state["buzzer"] == "on")
        assert self.state().child_lock == (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().target_humidity == self.device.start_state["auto"]
        assert self.state().fan_speed == FanSpeed(self.device.start_state["fan_speed"])
        assert self.state().tank_full == (self.device.start_state["tank_full"] == "on")
        assert self.state().compressor_status == (
            self.device.start_state["compressor_status"] == "on"
        )
        assert self.state().defrost_status == (
            self.device.start_state["defrost_status"] == "on"
        )
        assert self.state().fan_st == self.device.start_state["fan_st"]
        assert self.state().alarm == self.device.start_state["alarm"]

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.On)
        assert mode() == OperationMode.On

        self.device.set_mode(OperationMode.Auto)
        assert mode() == OperationMode.Auto

        self.device.set_mode(OperationMode.DryCloth)
        assert mode() == OperationMode.DryCloth

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_status_without_temperature(self):
        self.device._reset_state()
        self.device.state["temp"] = None

        assert self.state().temperature is None

    def test_set_target_humidity(self):
        def target_humidity():
            return self.device.status().target_humidity

        self.device.set_target_humidity(40)
        assert target_humidity() == 40
        self.device.set_target_humidity(50)
        assert target_humidity() == 50
        self.device.set_target_humidity(60)
        assert target_humidity() == 60

        with pytest.raises(AirDehumidifierException):
            self.device.set_target_humidity(-1)

        with pytest.raises(AirDehumidifierException):
            self.device.set_target_humidity(30)

        with pytest.raises(AirDehumidifierException):
            self.device.set_target_humidity(70)

        with pytest.raises(AirDehumidifierException):
            self.device.set_target_humidity(110)

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False



import logging
from collections import defaultdict
from typing import Any, Dict

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)


class CeilException(DeviceException):
    pass


class CeilStatus(DeviceStatus):
    """Container for status reports from Xiaomi Philips LED Ceiling Lamp."""

    def __init__(self, data: Dict[str, Any]) -> None:
        # {'power': 'off', 'bright': 0, 'snm': 4, 'dv': 0,
        #  'cctsw': [[0, 3], [0, 2], [0, 1]], 'bl': 1,
        #  'mb': 1, 'ac': 1, 'mssw': 1, 'cct': 99}

        # NOTE: Only 8 properties can be requested at the same time
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.power == "on"

    @property
    def brightness(self) -> int:
        """Current brightness."""
        return self.data["bright"]

    @property
    def scene(self) -> int:
        """Current fixed scene (brightness & colortemp)."""
        return self.data["snm"]

    @property
    def delay_off_countdown(self) -> int:
        """Countdown until turning off in seconds."""
        return self.data["dv"]

    @property
    def color_temperature(self) -> int:
        """Current color temperature."""
        return self.data["cct"]

    @property
    def smart_night_light(self) -> bool:
        """Smart night mode state."""
        return self.data["bl"] == 1

    @property
    def automatic_color_temperature(self) -> bool:
        """Automatic color temperature state."""
        return self.data["ac"] == 1


class Ceil(Device):
    """Main class representing Xiaomi Philips LED Ceiling Lamp."""

    # TODO: - Auto On/Off Not Supported
    #       - Adjust Scenes with Wall Switch Not Supported

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Brightness: {result.brightness}\n"
            "Color temperature: {result.color_temperature}\n"
            "Scene: {result.scene}\n"
            "Delayed turn off: {result.delay_off_countdown}\n"
            "Smart night light: {result.smart_night_light}\n"
            "Automatic color temperature: {result.automatic_color_temperature}\n",
        )
    )
    def status(self) -> CeilStatus:
        """Retrieve properties."""
        properties = ["power", "bright", "cct", "snm", "dv", "bl", "ac"]
        values = self.get_properties(properties)

        return CeilStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering on"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting brightness to {level}"),
    )
    def set_brightness(self, level: int):
        """Set brightness level."""
        if level < 1 or level > 100:
            raise CeilException("Invalid brightness: %s" % level)

        return self.send("set_bright", [level])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting color temperature to {level}"),
    )
    def set_color_temperature(self, level: int):
        """Set Correlated Color Temperature."""
        if level < 1 or level > 100:
            raise CeilException("Invalid color temperature: %s" % level)

        return self.send("set_cct", [level])

    @command(
        click.argument("brightness", type=int),
        click.argument("cct", type=int),
        default_output=format_output(
            "Setting brightness to {brightness} and color temperature to {cct}"
        ),
    )
    def set_brightness_and_color_temperature(self, brightness: int, cct: int):
        """Set brightness level and the correlated color temperature."""
        if brightness < 1 or brightness > 100:
            raise CeilException("Invalid brightness: %s" % brightness)

        if cct < 1 or cct > 100:
            raise CeilException("Invalid color temperature: %s" % cct)

        return self.send("set_bricct", [brightness, cct])

    @command(
        click.argument("seconds", type=int),
        default_output=format_output("Setting delayed turn off to {seconds} seconds"),
    )
    def delay_off(self, seconds: int):
        """Turn off delay in seconds."""

        if seconds < 1:
            raise CeilException("Invalid value for a delayed turn off: %s" % seconds)

        return self.send("delay_off", [seconds])

    @command(
        click.argument("number", type=int),
        default_output=format_output("Setting fixed scene to {number}"),
    )
    def set_scene(self, number: int):
        """Set a fixed scene (1-4)."""
        if number < 1 or number > 4:
            raise CeilException("Invalid fixed scene number: %s" % number)

        return self.send("apply_fixed_scene", [number])

    @command(default_output=format_output("Turning on smart night light"))
    def smart_night_light_on(self):
        """Smart Night Light On."""
        return self.send("enable_bl", [1])

    @command(default_output=format_output("Turning off smart night light"))
    def smart_night_light_off(self):
        """Smart Night Light off."""
        return self.send("enable_bl", [0])

    @command(default_output=format_output("Turning on automatic color temperature"))
    def automatic_color_temperature_on(self):
        """Automatic color temperature on."""
        return self.send("enable_ac", [1])

    @command(default_output=format_output("Turning off automatic color temperature"))
    def automatic_color_temperature_off(self):
        """Automatic color temperature off."""
        return self.send("enable_ac", [0])



"""Xiaomi Gateway subdevice base class."""

# flake8: noqa
from .light import LightBulb
from .sensor import Vibration
from .switch import Switch

from .subdevice import SubDevice, SubDeviceInfo  # isort:skip



from datetime import timedelta
from unittest import TestCase

import pytest

from miio import Walkingpad
from miio.walkingpad import (
    OperationMode,
    OperationSensitivity,
    WalkingpadException,
    WalkingpadStatus,
)

from .dummies import DummyDevice


class DummyWalkingpad(DummyDevice, Walkingpad):
    def _get_state(self, props):
        """Return wanted properties."""

        # Overriding here to deal with case of 'all' being requested

        if props[0] == "all":
            return self.state[props[0]]

        return [self.state[x] for x in props if x in self.state]

    def _set_state(self, var, value):
        """Set a state of a variable, the value is expected to be an array with length
        of 1."""

        # Overriding here to deal with case of 'all' being set

        if var == "all":
            self.state[var] = value
        else:
            self.state[var] = value.pop(0)

    def __init__(self, *args, **kwargs):
        self.state = {
            "power": "on",
            "mode": OperationMode.Manual,
            "time": 1387,
            "step": 2117,
            "sensitivity": OperationSensitivity.Low,
            "dist": 1150,
            "sp": 3.15,
            "cal": 71710,
            "start_speed": 3.1,
            "all": [
                "mode:" + str(OperationMode.Manual.value),
                "time:1387",
                "sp:3.15",
                "dist:1150",
                "cal:71710",
                "step:2117",
            ],
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_speed": lambda x: (
                self._set_state(
                    "all",
                    [
                        "mode:1",
                        "time:1387",
                        "sp:" + str(x[0]),
                        "dist:1150",
                        "cal:71710",
                        "step:2117",
                    ],
                ),
                self._set_state("sp", x),
            ),
            "set_step": lambda x: self._set_state("step", x),
            "set_sensitivity": lambda x: self._set_state("sensitivity", x),
            "set_start_speed": lambda x: self._set_state("start_speed", x),
            "set_time": lambda x: self._set_state("time", x),
            "set_distance": lambda x: self._set_state("dist", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def walkingpad(request):
    request.cls.device = DummyWalkingpad()


@pytest.mark.usefixtures("walkingpad")
class TestWalkingpad(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(WalkingpadStatus(self.device.start_state))
        assert self.is_on() is True
        assert self.state().power == self.device.start_state["power"]
        assert self.state().mode == self.device.start_state["mode"]
        assert self.state().speed == self.device.start_state["sp"]
        assert self.state().step_count == self.device.start_state["step"]
        assert self.state().distance == self.device.start_state["dist"]
        assert self.state().sensitivity == self.device.start_state["sensitivity"]
        assert self.state().walking_time == timedelta(
            seconds=self.device.start_state["time"]
        )

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Auto)
        assert mode() == OperationMode.Auto

        self.device.set_mode(OperationMode.Manual)
        assert mode() == OperationMode.Manual

        with pytest.raises(WalkingpadException):
            self.device.set_mode(-1)

        with pytest.raises(WalkingpadException):
            self.device.set_mode(3)

        with pytest.raises(WalkingpadException):
            self.device.set_mode("blah")

    def test_set_speed(self):
        def speed():
            return self.device.status().speed

        self.device.on()
        self.device.set_speed(3.055)
        assert speed() == 3.055

        with pytest.raises(WalkingpadException):
            self.device.set_speed(7.6)

        with pytest.raises(WalkingpadException):
            self.device.set_speed(-1)

        with pytest.raises(WalkingpadException):
            self.device.set_speed("blah")

        with pytest.raises(WalkingpadException):
            self.device.off()
            self.device.set_speed(3.4)

    def test_set_start_speed(self):
        def speed():
            return self.device.status().start_speed

        self.device.on()

        self.device.set_start_speed(3.055)
        assert speed() == 3.055

        with pytest.raises(WalkingpadException):
            self.device.set_start_speed(7.6)

        with pytest.raises(WalkingpadException):
            self.device.set_start_speed(-1)

        with pytest.raises(WalkingpadException):
            self.device.set_start_speed("blah")

        with pytest.raises(WalkingpadException):
            self.device.off()
            self.device.set_start_speed(3.4)

    def test_set_sensitivity(self):
        def sensitivity():
            return self.device.status().sensitivity

        self.device.set_sensitivity(OperationSensitivity.High)
        assert sensitivity() == OperationSensitivity.High

        self.device.set_sensitivity(OperationSensitivity.Medium)
        assert sensitivity() == OperationSensitivity.Medium

        with pytest.raises(WalkingpadException):
            self.device.set_sensitivity(-1)

        with pytest.raises(WalkingpadException):
            self.device.set_sensitivity(99)

        with pytest.raises(WalkingpadException):
            self.device.set_sensitivity("blah")



import logging
from collections import defaultdict
from typing import Optional

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_AIRQUALITYMONITOR_V1 = "zhimi.airmonitor.v1"
MODEL_AIRQUALITYMONITOR_B1 = "cgllc.airmonitor.b1"
MODEL_AIRQUALITYMONITOR_S1 = "cgllc.airmonitor.s1"

AVAILABLE_PROPERTIES_COMMON = [
    "power",
    "aqi",
    "battery",
    "usb_state",
    "time_state",
    "night_state",
    "night_beg_time",
    "night_end_time",
    "sensor_state",
]

AVAILABLE_PROPERTIES_B1 = ["co2e", "humidity", "pm25", "temperature", "tvoc"]

AVAILABLE_PROPERTIES_S1 = ["battery", "co2", "humidity", "pm25", "temperature", "tvoc"]

AVAILABLE_PROPERTIES = {
    MODEL_AIRQUALITYMONITOR_V1: AVAILABLE_PROPERTIES_COMMON,
    MODEL_AIRQUALITYMONITOR_B1: AVAILABLE_PROPERTIES_B1,
    MODEL_AIRQUALITYMONITOR_S1: AVAILABLE_PROPERTIES_S1,
}


class AirQualityMonitorException(DeviceException):
    pass


class AirQualityMonitorStatus(DeviceStatus):
    """Container of air quality monitor status."""

    def __init__(self, data):
        """Response of a Xiaomi Air Quality Monitor (zhimi.airmonitor.v1):

        {'power': 'on', 'aqi': 34, 'battery': 100, 'usb_state': 'off', 'time_state': 'on'}

        Response of a Xiaomi Air Quality Monitor (cgllc.airmonitor.b1):

        {'co2e': 1466, 'humidity': 59.79999923706055, 'pm25': 2, 'temperature': 19.799999237060547,
         'temperature_unit': 'c', 'tvoc': 1.3948699235916138, 'tvoc_unit': 'mg_m3'}

        Response of a Xiaomi Air Quality Monitor (cgllc.airmonitor.s1):

        {'battery': 100, 'co2': 695, 'humidity': 62.1, 'pm25': 19.4, 'temperature': 27.4,
         'tvoc': 254}
        """
        self.data = data

    @property
    def power(self) -> Optional[str]:
        """Current power state."""
        return self.data.get("power", None)

    @property
    def is_on(self) -> bool:
        """Return True if the device is turned on."""
        return self.power == "on"

    @property
    def usb_power(self) -> Optional[bool]:
        """Return True if the device's usb is on."""
        if "usb_state" in self.data and self.data["usb_state"] is not None:
            return self.data["usb_state"] == "on"
        return None

    @property
    def aqi(self) -> Optional[int]:
        """Air quality index value (0..600)."""
        return self.data.get("aqi", None)

    @property
    def battery(self) -> Optional[int]:
        """Current battery level (0..100)."""
        return self.data.get("battery", None)

    @property
    def display_clock(self) -> Optional[bool]:
        """Display a clock instead the AQI."""
        if "time_state" in self.data and self.data["time_state"] is not None:
            return self.data["time_state"] == "on"
        return None

    @property
    def night_mode(self) -> Optional[bool]:
        """Return True if the night mode is on."""
        if "night_state" in self.data and self.data["night_state"] is not None:
            return self.data["night_state"] == "on"
        return None

    @property
    def night_time_begin(self) -> Optional[str]:
        """Return the begin of the night time."""
        return self.data.get("night_beg_time", None)

    @property
    def night_time_end(self) -> Optional[str]:
        """Return the end of the night time."""
        return self.data.get("night_end_time", None)

    @property
    def sensor_state(self) -> Optional[str]:
        """Sensor state."""
        return self.data.get("sensor_state", None)

    @property
    def co2(self) -> Optional[int]:
        """Return co2 value (400...9999ppm)."""
        return self.data.get("co2", None)

    @property
    def co2e(self) -> Optional[int]:
        """Return co2e value (400...9999ppm)."""
        return self.data.get("co2e", None)

    @property
    def humidity(self) -> Optional[float]:
        """Return humidity value (0...100%)."""
        return self.data.get("humidity", None)

    @property
    def pm25(self) -> Optional[float]:
        """Return pm2.5 value (0...999μg/m³)."""
        return self.data.get("pm25", None)

    @property
    def temperature(self) -> Optional[float]:
        """Return temperature value (-10...50°C)."""
        return self.data.get("temperature", None)

    @property
    def tvoc(self) -> Optional[int]:
        """Return tvoc value."""
        return self.data.get("tvoc", None)


class AirQualityMonitor(Device):
    """Xiaomi PM2.5 Air Quality Monitor."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_AIRQUALITYMONITOR_V1,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        elif model is not None:
            self.model = MODEL_AIRQUALITYMONITOR_V1
            _LOGGER.error(
                "Device model %s unsupported. Falling back to %s.", model, self.model
            )
        else:
            # Force autodetection.
            self.model = None

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "USB power: {result.usb_power}\n"
            "Battery: {result.battery}\n"
            "AQI: {result.aqi}\n"
            "Temperature: {result.temperature}\n"
            "Humidity: {result.humidity}\n"
            "CO2: {result.co2}\n"
            "CO2e: {result.co2e}\n"
            "PM2.5: {result.pm25}\n"
            "TVOC: {result.tvoc}\n"
            "Display clock: {result.display_clock}\n",
        )
    )
    def status(self) -> AirQualityMonitorStatus:
        """Return device status."""

        if self.model is None:
            info = self.info()
            self.model = info.model

        properties = AVAILABLE_PROPERTIES[self.model]

        if self.model == MODEL_AIRQUALITYMONITOR_B1:
            values = self.send("get_air_data")
        else:
            values = self.send("get_prop", properties)

        properties_count = len(properties)
        values_count = len(values)
        if properties_count != values_count:
            _LOGGER.debug(
                "Count (%s) of requested properties does not match the "
                "count (%s) of received values.",
                properties_count,
                values_count,
            )

        if (
            self.model == MODEL_AIRQUALITYMONITOR_S1
            or self.model == MODEL_AIRQUALITYMONITOR_B1
        ):
            return AirQualityMonitorStatus(defaultdict(lambda: None, values))
        else:
            return AirQualityMonitorStatus(
                defaultdict(lambda: None, zip(properties, values))
            )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("display_clock", type=bool),
        default_output=format_output(
            lambda led: "Turning on display clock"
            if led
            else "Turning off display clock"
        ),
    )
    def set_display_clock(self, display_clock: bool):
        """Enable/disable displaying a clock instead the AQI."""
        if display_clock:
            self.send("set_time_state", ["on"])
        else:
            self.send("set_time_state", ["off"])

    @command(
        click.argument("auto_close", type=bool),
        default_output=format_output(
            lambda led: "Turning on auto close" if led else "Turning off auto close"
        ),
    )
    def set_auto_close(self, auto_close: bool):
        """Purpose unknown."""
        if auto_close:
            self.send("set_auto_close", ["on"])
        else:
            self.send("set_auto_close", ["off"])

    @command(
        click.argument("night_mode", type=bool),
        default_output=format_output(
            lambda led: "Turning on night mode" if led else "Turning off night mode"
        ),
    )
    def set_night_mode(self, night_mode: bool):
        """Decrease the brightness of the display."""
        if night_mode:
            self.send("set_night_state", ["on"])
        else:
            self.send("set_night_state", ["off"])

    @command(
        click.argument("begin_hour", type=int),
        click.argument("begin_minute", type=int),
        click.argument("end_hour", type=int),
        click.argument("end_minute", type=int),
        default_output=format_output(
            "Setting night time to {begin_hour}:{begin_minute} - {end_hour}:{end_minute}"
        ),
    )
    def set_night_time(
        self, begin_hour: int, begin_minute: int, end_hour: int, end_minute: int
    ):
        """Enable night mode daily at bedtime."""
        begin = begin_hour * 3600 + begin_minute * 60
        end = end_hour * 3600 + end_minute * 60

        if begin < 0 or begin > 86399 or end < 0 or end > 86399:
            AirQualityMonitorException("Begin or/and end time invalid.")

        self.send("set_night_time", [begin, end])



from unittest import TestCase

import pytest

from miio import AirPurifierMiot
from miio.airfilter_util import FilterType
from miio.airpurifier_miot import AirPurifierMiotException, LedBrightness, OperationMode

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "power": True,
    "aqi": 10,
    "average_aqi": 8,
    "humidity": 62,
    "temperature": 18.599999,
    "fan_level": 2,
    "mode": 0,
    "led": True,
    "led_brightness": 1,
    "buzzer": False,
    "buzzer_volume": 0,
    "child_lock": False,
    "favorite_level": 10,
    "filter_life_remaining": 80,
    "filter_hours_used": 682,
    "use_time": 2457000,
    "purify_volume": 25262,
    "motor_speed": 354,
    "filter_rfid_product_id": "0:0:41:30",
    "filter_rfid_tag": "10:20:30:40:50:60:7",
    "button_pressed": "power",
}


class DummyAirPurifierMiot(DummyMiotDevice, AirPurifierMiot):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_led": lambda x: self._set_state("led", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_level_favorite": lambda x: self._set_state("favorite_level", x),
            "set_led_b": lambda x: self._set_state("led_b", x),
            "set_volume": lambda x: self._set_state("volume", x),
            "set_act_sleep": lambda x: self._set_state("act_sleep", x),
            "reset_filter1": lambda x: (
                self._set_state("f1_hour_used", [0]),
                self._set_state("filter1_life", [100]),
            ),
            "set_act_det": lambda x: self._set_state("act_det", x),
            "set_app_extra": lambda x: self._set_state("app_extra", x),
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def airpurifier(request):
    request.cls.device = DummyAirPurifierMiot()


@pytest.mark.usefixtures("airpurifier")
class TestAirPurifier(TestCase):
    def test_on(self):
        self.device.off()  # ensure off
        assert self.device.status().is_on is False

        self.device.on()
        assert self.device.status().is_on is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.device.status().is_on is True

        self.device.off()
        assert self.device.status().is_on is False

    def test_status(self):
        status = self.device.status()
        assert status.is_on is _INITIAL_STATE["power"]
        assert status.aqi == _INITIAL_STATE["aqi"]
        assert status.average_aqi == _INITIAL_STATE["average_aqi"]
        assert status.humidity == _INITIAL_STATE["humidity"]
        assert status.temperature == 18.6
        assert status.fan_level == _INITIAL_STATE["fan_level"]
        assert status.mode == OperationMode(_INITIAL_STATE["mode"])
        assert status.led == _INITIAL_STATE["led"]
        assert status.led_brightness == LedBrightness(_INITIAL_STATE["led_brightness"])
        assert status.buzzer == _INITIAL_STATE["buzzer"]
        assert status.child_lock == _INITIAL_STATE["child_lock"]
        assert status.favorite_level == _INITIAL_STATE["favorite_level"]
        assert status.filter_life_remaining == _INITIAL_STATE["filter_life_remaining"]
        assert status.filter_hours_used == _INITIAL_STATE["filter_hours_used"]
        assert status.use_time == _INITIAL_STATE["use_time"]
        assert status.purify_volume == _INITIAL_STATE["purify_volume"]
        assert status.motor_speed == _INITIAL_STATE["motor_speed"]
        assert status.filter_rfid_product_id == _INITIAL_STATE["filter_rfid_product_id"]
        assert status.filter_type == FilterType.AntiBacterial

    def test_set_fan_level(self):
        def fan_level():
            return self.device.status().fan_level

        self.device.set_fan_level(1)
        assert fan_level() == 1
        self.device.set_fan_level(2)
        assert fan_level() == 2
        self.device.set_fan_level(3)
        assert fan_level() == 3

        with pytest.raises(AirPurifierMiotException):
            self.device.set_fan_level(0)

        with pytest.raises(AirPurifierMiotException):
            self.device.set_fan_level(4)

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Auto)
        assert mode() == OperationMode.Auto

        self.device.set_mode(OperationMode.Silent)
        assert mode() == OperationMode.Silent

        self.device.set_mode(OperationMode.Favorite)
        assert mode() == OperationMode.Favorite

        self.device.set_mode(OperationMode.Fan)
        assert mode() == OperationMode.Fan

    def test_set_favorite_level(self):
        def favorite_level():
            return self.device.status().favorite_level

        self.device.set_favorite_level(0)
        assert favorite_level() == 0
        self.device.set_favorite_level(6)
        assert favorite_level() == 6
        self.device.set_favorite_level(14)
        assert favorite_level() == 14

        with pytest.raises(AirPurifierMiotException):
            self.device.set_favorite_level(-1)

        with pytest.raises(AirPurifierMiotException):
            self.device.set_favorite_level(15)

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False



import math

import pytest

from miio import Device
from miio.exceptions import DeviceInfoUnavailableException, PayloadDecodeException


@pytest.mark.parametrize("max_properties", [None, 1, 15])
def test_get_properties_splitting(mocker, max_properties):
    properties = [i for i in range(20)]

    send = mocker.patch("miio.Device.send")
    d = Device("127.0.0.1", "68ffffffffffffffffffffffffffffff")
    d.get_properties(properties, max_properties=max_properties)

    if max_properties is None:
        max_properties = len(properties)
    assert send.call_count == math.ceil(len(properties) / max_properties)


def test_default_timeout_and_retry(mocker):
    send = mocker.patch("miio.miioprotocol.MiIOProtocol.send")
    d = Device("127.0.0.1", "68ffffffffffffffffffffffffffffff")
    assert d._protocol._timeout == 5
    d.send(command="fake_command", parameters=[])
    send.assert_called_with("fake_command", [], 3, extra_parameters=None)


def test_timeout_retry(mocker):
    send = mocker.patch("miio.miioprotocol.MiIOProtocol.send")
    d = Device("127.0.0.1", "68ffffffffffffffffffffffffffffff", timeout=4)
    assert d._protocol._timeout == 4
    d.send("fake_command", [], 1)
    send.assert_called_with("fake_command", [], 1, extra_parameters=None)
    d.send("fake_command", [])
    send.assert_called_with("fake_command", [], 3, extra_parameters=None)

    class CustomDevice(Device):
        retry_count = 5
        timeout = 1

    d2 = CustomDevice("127.0.0.1", "68ffffffffffffffffffffffffffffff")
    assert d2._protocol._timeout == 1
    d2.send("fake_command", [])
    send.assert_called_with("fake_command", [], 5, extra_parameters=None)


def test_unavailable_device_info_raises(mocker):
    send = mocker.patch("miio.Device.send", side_effect=PayloadDecodeException)
    d = Device("127.0.0.1", "68ffffffffffffffffffffffffffffff")

    with pytest.raises(DeviceInfoUnavailableException):
        d.info()

    assert send.call_count == 1



import enum
import logging
from typing import Any, Dict, Optional

import click

from .airfilter_util import FilterType, FilterTypeUtil
from .click_common import EnumType, command, format_output
from .exceptions import DeviceException
from .miot_device import DeviceStatus, MiotDevice

_LOGGER = logging.getLogger(__name__)
_MAPPING = {
    # Air Purifier (siid=2)
    "power": {"siid": 2, "piid": 2},
    "fan_level": {"siid": 2, "piid": 4},
    "mode": {"siid": 2, "piid": 5},
    # Environment (siid=3)
    "humidity": {"siid": 3, "piid": 7},
    "temperature": {"siid": 3, "piid": 8},
    "aqi": {"siid": 3, "piid": 6},
    # Filter (siid=4)
    "filter_life_remaining": {"siid": 4, "piid": 3},
    "filter_hours_used": {"siid": 4, "piid": 5},
    # Alarm (siid=5)
    "buzzer": {"siid": 5, "piid": 1},
    "buzzer_volume": {"siid": 5, "piid": 2},
    # Indicator Light (siid=6)
    "led_brightness": {"siid": 6, "piid": 1},
    "led": {"siid": 6, "piid": 6},
    # Physical Control Locked (siid=7)
    "child_lock": {"siid": 7, "piid": 1},
    # Motor Speed (siid=10)
    "favorite_level": {"siid": 10, "piid": 10},
    "favorite_rpm": {"siid": 10, "piid": 7},
    "motor_speed": {"siid": 10, "piid": 8},
    # Use time (siid=12)
    "use_time": {"siid": 12, "piid": 1},
    # AQI (siid=13)
    "purify_volume": {"siid": 13, "piid": 1},
    "average_aqi": {"siid": 13, "piid": 2},
    # RFID (siid=14)
    "filter_rfid_tag": {"siid": 14, "piid": 1},
    "filter_rfid_product_id": {"siid": 14, "piid": 3},
    # Other (siid=15)
    "app_extra": {"siid": 15, "piid": 1},
}

# https://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:air-purifier:0000A007:zhimi-mb4:2
_MODEL_AIRPURIFIER_MB4 = {
    # Air Purifier
    "power": {"siid": 2, "piid": 1},
    "mode": {"siid": 2, "piid": 4},
    # Environment
    "aqi": {"siid": 3, "piid": 4},
    # Filter
    "filter_life_remaining": {"siid": 4, "piid": 1},
    "filter_hours_used": {"siid": 4, "piid": 3},
    # Alarm
    "buzzer": {"siid": 6, "piid": 1},
    # Screen
    "led_brightness_level": {"siid": 7, "piid": 2},
    # Physical Control Locked
    "child_lock": {"siid": 8, "piid": 1},
    # custom-service
    "motor_speed": {"siid": 9, "piid": 1},
    "favorite_rpm": {"siid": 9, "piid": 3},
}


class AirPurifierMiotException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Unknown = -1
    Auto = 0
    Silent = 1
    Favorite = 2
    Fan = 3


class LedBrightness(enum.Enum):
    Bright = 0
    Dim = 1
    Off = 2


class BasicAirPurifierMiotStatus(DeviceStatus):
    """Container for status reports from the air purifier."""

    def __init__(self, data: Dict[str, Any]) -> None:
        self.filter_type_util = FilterTypeUtil()
        self.data = data

    @property
    def is_on(self) -> bool:
        """Return True if device is on."""
        return self.data["power"]

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.is_on else "off"

    @property
    def aqi(self) -> int:
        """Air quality index."""
        return self.data["aqi"]

    @property
    def mode(self) -> OperationMode:
        """Current operation mode."""
        mode = self.data["mode"]
        try:
            return OperationMode(mode)
        except ValueError:
            _LOGGER.debug("Unknown mode: %s", mode)
            return OperationMode.Unknown

    @property
    def buzzer(self) -> Optional[bool]:
        """Return True if buzzer is on."""
        if self.data["buzzer"] is not None:
            return self.data["buzzer"]

        return None

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["child_lock"]

    @property
    def filter_life_remaining(self) -> int:
        """Time until the filter should be changed."""
        return self.data["filter_life_remaining"]

    @property
    def filter_hours_used(self) -> int:
        """How long the filter has been in use."""
        return self.data["filter_hours_used"]

    @property
    def motor_speed(self) -> int:
        """Speed of the motor."""
        return self.data["motor_speed"]

    @property
    def favorite_rpm(self) -> Optional[int]:
        """Return favorite rpm level."""
        return self.data.get("favorite_rpm")


class AirPurifierMiotStatus(BasicAirPurifierMiotStatus):
    """Container for status reports from the air purifier.

    Mi Air Purifier 3/3H (zhimi.airpurifier.mb3) response (MIoT format)

    [
        {'did': 'power', 'siid': 2, 'piid': 2, 'code': 0, 'value': True},
        {'did': 'fan_level', 'siid': 2, 'piid': 4, 'code': 0, 'value': 1},
        {'did': 'mode', 'siid': 2, 'piid': 5, 'code': 0, 'value': 2},
        {'did': 'humidity', 'siid': 3, 'piid': 7, 'code': 0, 'value': 38},
        {'did': 'temperature', 'siid': 3, 'piid': 8, 'code': 0, 'value': 22.299999},
        {'did': 'aqi', 'siid': 3, 'piid': 6, 'code': 0, 'value': 2},
        {'did': 'filter_life_remaining', 'siid': 4, 'piid': 3, 'code': 0, 'value': 45},
        {'did': 'filter_hours_used', 'siid': 4, 'piid': 5, 'code': 0, 'value': 1915},
        {'did': 'buzzer', 'siid': 5, 'piid': 1, 'code': 0, 'value': False},
        {'did': 'buzzer_volume', 'siid': 5, 'piid': 2, 'code': -4001},
        {'did': 'led_brightness', 'siid': 6, 'piid': 1, 'code': 0, 'value': 1},
        {'did': 'led', 'siid': 6, 'piid': 6, 'code': 0, 'value': True},
        {'did': 'child_lock', 'siid': 7, 'piid': 1, 'code': 0, 'value': False},
        {'did': 'favorite_level', 'siid': 10, 'piid': 10, 'code': 0, 'value': 2},
        {'did': 'favorite_rpm', 'siid': 10, 'piid': 7, 'code': 0, 'value': 770},
        {'did': 'motor_speed', 'siid': 10, 'piid': 8, 'code': 0, 'value': 769},
        {'did': 'use_time', 'siid': 12, 'piid': 1, 'code': 0, 'value': 6895800},
        {'did': 'purify_volume', 'siid': 13, 'piid': 1, 'code': 0, 'value': 222564},
        {'did': 'average_aqi', 'siid': 13, 'piid': 2, 'code': 0, 'value': 2},
        {'did': 'filter_rfid_tag', 'siid': 14, 'piid': 1, 'code': 0, 'value': '81:6b:3f:32:84:4b:4'},
        {'did': 'filter_rfid_product_id', 'siid': 14, 'piid': 3, 'code': 0, 'value': '0:0:31:31'},
        {'did': 'app_extra', 'siid': 15, 'piid': 1, 'code': 0, 'value': 0}
    ]
    """

    @property
    def average_aqi(self) -> int:
        """Average of the air quality index."""
        return self.data["average_aqi"]

    @property
    def humidity(self) -> int:
        """Current humidity."""
        return self.data["humidity"]

    @property
    def temperature(self) -> Optional[float]:
        """Current temperature, if available."""
        if self.data["temperature"] is not None:
            return round(self.data["temperature"], 1)

        return None

    @property
    def fan_level(self) -> int:
        """Current fan level."""
        return self.data["fan_level"]

    @property
    def led(self) -> bool:
        """Return True if LED is on."""
        return self.data["led"]

    @property
    def led_brightness(self) -> Optional[LedBrightness]:
        """Brightness of the LED."""
        if self.data["led_brightness"] is not None:
            try:
                return LedBrightness(self.data["led_brightness"])
            except ValueError:
                return None

        return None

    @property
    def buzzer_volume(self) -> Optional[int]:
        """Return buzzer volume."""
        if self.data["buzzer_volume"] is not None:
            return self.data["buzzer_volume"]

        return None

    @property
    def favorite_level(self) -> int:
        """Return favorite level, which is used if the mode is ``favorite``."""
        # Favorite level used when the mode is `favorite`.
        return self.data["favorite_level"]

    @property
    def use_time(self) -> int:
        """How long the device has been active in seconds."""
        return self.data["use_time"]

    @property
    def purify_volume(self) -> int:
        """The volume of purified air in cubic meter."""
        return self.data["purify_volume"]

    @property
    def filter_rfid_product_id(self) -> Optional[str]:
        """RFID product ID of installed filter."""
        return self.data["filter_rfid_product_id"]

    @property
    def filter_rfid_tag(self) -> Optional[str]:
        """RFID tag ID of installed filter."""
        return self.data["filter_rfid_tag"]

    @property
    def filter_type(self) -> Optional[FilterType]:
        """Type of installed filter."""
        return self.filter_type_util.determine_filter_type(
            self.filter_rfid_tag, self.filter_rfid_product_id
        )


class AirPurifierMB4Status(BasicAirPurifierMiotStatus):
    """
    Container for status reports from the  Mi Air Purifier 3C (zhimi.airpurifier.mb4).

    {
        'power': True,
        'mode': 1,
        'aqi': 2,
        'filter_life_remaining': 97,
        'filter_hours_used': 100,
        'buzzer': True,
        'led_brightness_level': 8,
        'child_lock': False,
        'motor_speed': 392,
        'favorite_rpm': 500
    }

    Response (MIoT format)

    [
        {'did': 'power', 'siid': 2, 'piid': 1, 'code': 0, 'value': True},
        {'did': 'mode', 'siid': 2, 'piid': 4, 'code': 0, 'value': 1},
        {'did': 'aqi', 'siid': 3, 'piid': 4, 'code': 0, 'value': 3},
        {'did': 'filter_life_remaining', 'siid': 4, 'piid': 1, 'code': 0, 'value': 97},
        {'did': 'filter_hours_used', 'siid': 4, 'piid': 3, 'code': 0, 'value': 100},
        {'did': 'buzzer', 'siid': 6, 'piid': 1, 'code': 0, 'value': True},
        {'did': 'led_brightness_level', 'siid': 7, 'piid': 2, 'code': 0, 'value': 8},
        {'did': 'child_lock', 'siid': 8, 'piid': 1, 'code': 0, 'value': False},
        {'did': 'motor_speed', 'siid': 9, 'piid': 1, 'code': 0, 'value': 388},
        {'did': 'favorite_rpm', 'siid': 9, 'piid': 3, 'code': 0, 'value': 500}
    ]

    """

    @property
    def led_brightness_level(self) -> int:
        """Return brightness level."""
        return self.data["led_brightness_level"]


class BasicAirPurifierMiot(MiotDevice):
    """Main class representing the air purifier which uses MIoT protocol."""

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.set_property("power", True)

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.set_property("power", False)

    @command(
        click.argument("rpm", type=int),
        default_output=format_output("Setting favorite motor speed '{rpm}' rpm"),
    )
    def set_favorite_rpm(self, rpm: int):
        """Set favorite motor speed."""
        # Note: documentation says the maximum is 2300, however, the purifier may return an error for rpm over 2200.
        if rpm < 300 or rpm > 2300 or rpm % 10 != 0:
            raise AirPurifierMiotException(
                "Invalid favorite motor speed: %s. Must be between 300 and 2300 and divisible by 10"
                % rpm
            )
        return self.set_property("favorite_rpm", rpm)

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        return self.set_property("mode", mode.value)

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        return self.set_property("buzzer", buzzer)

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.set_property("child_lock", lock)


class AirPurifierMiot(BasicAirPurifierMiot):
    """Main class representing the air purifier which uses MIoT protocol."""

    mapping = _MAPPING

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "AQI: {result.aqi} μg/m³\n"
            "Average AQI: {result.average_aqi} μg/m³\n"
            "Humidity: {result.humidity} %\n"
            "Temperature: {result.temperature} °C\n"
            "Fan Level: {result.fan_level}\n"
            "Mode: {result.mode}\n"
            "LED: {result.led}\n"
            "LED brightness: {result.led_brightness}\n"
            "Buzzer: {result.buzzer}\n"
            "Buzzer vol.: {result.buzzer_volume}\n"
            "Child lock: {result.child_lock}\n"
            "Favorite level: {result.favorite_level}\n"
            "Filter life remaining: {result.filter_life_remaining} %\n"
            "Filter hours used: {result.filter_hours_used}\n"
            "Use time: {result.use_time} s\n"
            "Purify volume: {result.purify_volume} m³\n"
            "Motor speed: {result.motor_speed} rpm\n"
            "Filter RFID product id: {result.filter_rfid_product_id}\n"
            "Filter RFID tag: {result.filter_rfid_tag}\n"
            "Filter type: {result.filter_type}\n",
        )
    )
    def status(self) -> AirPurifierMiotStatus:
        """Retrieve properties."""

        return AirPurifierMiotStatus(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting fan level to '{level}'"),
    )
    def set_fan_level(self, level: int):
        """Set fan level."""
        if level < 1 or level > 3:
            raise AirPurifierMiotException("Invalid fan level: %s" % level)
        return self.set_property("fan_level", level)

    @command(
        click.argument("volume", type=int),
        default_output=format_output("Setting sound volume to {volume}"),
    )
    def set_volume(self, volume: int):
        """Set buzzer volume."""
        if volume < 0 or volume > 100:
            raise AirPurifierMiotException(
                "Invalid volume: %s. Must be between 0 and 100" % volume
            )
        return self.set_property("buzzer_volume", volume)

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting favorite level to {level}"),
    )
    def set_favorite_level(self, level: int):
        """Set the favorite level used when the mode is `favorite`.

        Needs to be between 0 and 14.
        """
        if level < 0 or level > 14:
            raise AirPurifierMiotException("Invalid favorite level: %s" % level)

        return self.set_property("favorite_level", level)

    @command(
        click.argument("brightness", type=EnumType(LedBrightness)),
        default_output=format_output("Setting LED brightness to {brightness}"),
    )
    def set_led_brightness(self, brightness: LedBrightness):
        """Set led brightness."""
        return self.set_property("led_brightness", brightness.value)

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        return self.set_property("led", led)


class AirPurifierMB4(BasicAirPurifierMiot):
    """Main class representing the air purifier which uses MIoT protocol."""

    mapping = _MODEL_AIRPURIFIER_MB4

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "AQI: {result.aqi} μg/m³\n"
            "Mode: {result.mode}\n"
            "LED brightness level: {result.led_brightness_level}\n"
            "Buzzer: {result.buzzer}\n"
            "Child lock: {result.child_lock}\n"
            "Filter life remaining: {result.filter_life_remaining} %\n"
            "Filter hours used: {result.filter_hours_used}\n"
            "Motor speed: {result.motor_speed} rpm\n"
            "Favorite RPM: {result.favorite_rpm} rpm\n",
        )
    )
    def status(self) -> AirPurifierMB4Status:
        """Retrieve properties."""

        return AirPurifierMB4Status(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting LED brightness level to {level}"),
    )
    def set_led_brightness_level(self, level: int):
        """Set led brightness level (0..8)."""
        if level < 0 or level > 8:
            raise AirPurifierMiotException("Invalid brightness level: %s" % level)

        return self.set_property("led_brightness_level", level)



from unittest import TestCase

import pytest

from miio import AirHumidifierMjjsq
from miio.airhumidifier_mjjsq import (
    MODEL_HUMIDIFIER_JSQ1,
    AirHumidifierException,
    AirHumidifierStatus,
    OperationMode,
)

from .dummies import DummyDevice


class DummyAirHumidifierMjjsq(DummyDevice, AirHumidifierMjjsq):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_HUMIDIFIER_JSQ1
        self.state = {
            "Humidifier_Gear": 1,
            "Humidity_Value": 44,
            "HumiSet_Value": 11,
            "Led_State": 0,
            "OnOff_State": 1,
            "TemperatureValue": 21,
            "TipSound_State": 0,
            "waterstatus": 1,
            "watertankstatus": 1,
            "wet_and_protect": 1,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "Set_OnOff": lambda x: self._set_state("OnOff_State", x),
            "Set_HumidifierGears": lambda x: self._set_state("Humidifier_Gear", x),
            "SetLedState": lambda x: self._set_state("Led_State", x),
            "SetTipSound_Status": lambda x: self._set_state("TipSound_State", x),
            "Set_HumiValue": lambda x: self._set_state("HumiSet_Value", x),
            "Set_wet_and_protect": lambda x: self._set_state("wet_and_protect", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def airhumidifiermjjsq(request):
    request.cls.device = DummyAirHumidifierMjjsq()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airhumidifiermjjsq")
class TestAirHumidifierMjjsq(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(AirHumidifierStatus(self.device.start_state))
        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["TemperatureValue"]
        assert self.state().humidity == self.device.start_state["Humidity_Value"]
        assert self.state().mode == OperationMode(
            self.device.start_state["Humidifier_Gear"]
        )
        assert self.state().led is (self.device.start_state["Led_State"] == 1)
        assert self.state().buzzer is (self.device.start_state["TipSound_State"] == 1)
        assert self.state().target_humidity == self.device.start_state["HumiSet_Value"]
        assert self.state().no_water is (self.device.start_state["waterstatus"] == 0)
        assert self.state().water_tank_detached is (
            self.device.start_state["watertankstatus"] == 0
        )

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Low)
        assert mode() == OperationMode.Low

        self.device.set_mode(OperationMode.Medium)
        assert mode() == OperationMode.Medium

        self.device.set_mode(OperationMode.High)
        assert mode() == OperationMode.High

        self.device.set_mode(OperationMode.Humidity)
        assert mode() == OperationMode.Humidity

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_target_humidity(self):
        def target_humidity():
            return self.device.status().target_humidity

        self.device.set_target_humidity(0)
        assert target_humidity() == 0
        self.device.set_target_humidity(50)
        assert target_humidity() == 50
        self.device.set_target_humidity(99)
        assert target_humidity() == 99

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(-1)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(100)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(101)

    def test_set_wet_protection(self):
        def wet_protection():
            return self.device.status().wet_protection

        self.device.set_wet_protection(True)
        assert wet_protection() is True

        self.device.set_wet_protection(False)
        assert wet_protection() is False



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_POWER_STRIP_V1 = "qmi.powerstrip.v1"
MODEL_POWER_STRIP_V2 = "zimi.powerstrip.v2"

AVAILABLE_PROPERTIES = {
    MODEL_POWER_STRIP_V1: [
        "power",
        "temperature",
        "current",
        "mode",
        "power_consume_rate",
        "voltage",
        "power_factor",
        "elec_leakage",
    ],
    MODEL_POWER_STRIP_V2: [
        "power",
        "temperature",
        "current",
        "mode",
        "power_consume_rate",
        "wifi_led",
        "power_price",
    ],
}


class PowerStripException(DeviceException):
    pass


class PowerMode(enum.Enum):
    Eco = "green"
    Normal = "normal"


class PowerStripStatus(DeviceStatus):
    """Container for status reports from the power strip."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Supported device models: qmi.powerstrip.v1, zimi.powerstrip.v2.

        Response of a Power Strip 2 (zimi.powerstrip.v2):
        {'power','on', 'temperature': 48.7, 'current': 0.05, 'mode': None,
         'power_consume_rate': 4.09, 'wifi_led': 'on', 'power_price': 49}
        """
        self.data = data

    @property
    def power(self) -> str:
        """Current power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.power == "on"

    @property
    def temperature(self) -> float:
        """Current temperature."""
        return self.data["temperature"]

    @property
    def current(self) -> Optional[float]:
        """Current, if available.

        Meaning and voltage reference unknown.
        """
        if self.data["current"] is not None:
            return self.data["current"]
        return None

    @property
    def load_power(self) -> Optional[float]:
        """Current power load, if available."""
        if self.data["power_consume_rate"] is not None:
            return self.data["power_consume_rate"]
        return None

    @property
    def mode(self) -> Optional[PowerMode]:
        """Current operation mode, can be either green or normal."""
        if self.data["mode"] is not None:
            return PowerMode(self.data["mode"])
        return None

    @property
    def wifi_led(self) -> Optional[bool]:
        """True if the wifi led is turned on."""
        if "wifi_led" in self.data and self.data["wifi_led"] is not None:
            return self.data["wifi_led"] == "on"
        return None

    @property
    def power_price(self) -> Optional[int]:
        """The stored power price, if available."""
        if "power_price" in self.data and self.data["power_price"] is not None:
            return self.data["power_price"]
        return None

    @property
    def leakage_current(self) -> Optional[int]:
        """The leakage current, if available."""
        if "elec_leakage" in self.data and self.data["elec_leakage"] is not None:
            return self.data["elec_leakage"]
        return None

    @property
    def voltage(self) -> Optional[float]:
        """The voltage, if available."""
        if "voltage" in self.data and self.data["voltage"] is not None:
            return self.data["voltage"] / 100.0
        return None

    @property
    def power_factor(self) -> Optional[float]:
        """The power factor, if available."""
        if "power_factor" in self.data and self.data["power_factor"] is not None:
            return self.data["power_factor"]
        return None


class PowerStrip(Device):
    """Main class representing the smart power strip."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_POWER_STRIP_V1,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_POWER_STRIP_V1

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Temperature: {result.temperature} °C\n"
            "Voltage: {result.voltage} V\n"
            "Current: {result.current} A\n"
            "Load power: {result.load_power} W\n"
            "Power factor: {result.power_factor}\n"
            "Power price: {result.power_price}\n"
            "Leakage current: {result.leakage_current} A\n"
            "Mode: {result.mode}\n"
            "WiFi LED: {result.wifi_led}\n",
        )
    )
    def status(self) -> PowerStripStatus:
        """Retrieve properties."""
        properties = AVAILABLE_PROPERTIES[self.model]
        values = self.get_properties(properties)

        return PowerStripStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("mode", type=EnumType(PowerMode)),
        default_output=format_output("Setting mode to {mode}"),
    )
    def set_power_mode(self, mode: PowerMode):
        """Set the power mode."""

        # green, normal
        return self.send("set_power_mode", [mode.value])

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on WiFi LED" if led else "Turning off WiFi LED"
        ),
    )
    def set_wifi_led(self, led: bool):
        """Set the wifi led on/off."""
        if led:
            return self.send("set_wifi_led", ["on"])
        else:
            return self.send("set_wifi_led", ["off"])

    @command(
        click.argument("price", type=int),
        default_output=format_output("Setting power price to {price}"),
    )
    def set_power_price(self, price: int):
        """Set the power price."""
        if price < 0 or price > 999:
            raise PowerStripException("Invalid power price: %s" % price)

        return self.send("set_power_price", [price])

    @command(
        click.argument("power", type=bool),
        default_output=format_output(
            lambda led: "Turning on real-time power measurement"
            if led
            else "Turning off real-time power measurement"
        ),
    )
    def set_realtime_power(self, power: bool):
        """Set the realtime power on/off."""
        if power:
            return self.send("set_rt_power", [1])
        else:
            return self.send("set_rt_power", [0])



"""Xiaomi Gateway Light implementation."""

from typing import Tuple

import click

from ..click_common import command
from ..utils import brightness_and_color_to_int, int_to_brightness, int_to_rgb
from .gatewaydevice import GatewayDevice

color_map = {
    "red": (255, 0, 0),
    "green": (0, 255, 0),
    "blue": (0, 0, 255),
    "white": (255, 255, 255),
    "yellow": (255, 255, 0),
    "orange": (255, 165, 0),
    "aqua": (0, 255, 255),
    "olive": (128, 128, 0),
    "purple": (128, 0, 128),
}


class Light(GatewayDevice):
    """Light controls for the gateway.

    The gateway LEDs can be controlled using 'rgb' or 'night_light' methods. The
    'night_light' methods control the same light as the 'rgb' methods, but has a
    separate memory for brightness and color. Changing the 'rgb' light does not affect
    the stored state of the 'night_light', while changing the 'night_light' does effect
    the state of the 'rgb' light.
    """

    @command()
    def rgb_status(self):
        """Get current status of the light. Always represents the current status of the
        light as opposed to 'night_light_status'.

        Example:
           {"is_on": false, "brightness": 0, "rgb": (0, 0, 0)}
        """
        # Returns {"is_on": false, "brightness": 0, "rgb": (0, 0, 0)} when light is off
        state_int = self._gateway.send("get_rgb").pop()
        brightness = int_to_brightness(state_int)
        rgb = int_to_rgb(state_int)
        is_on = brightness > 0

        return {"is_on": is_on, "brightness": brightness, "rgb": rgb}

    @command()
    def night_light_status(self):
        """Get status of the night light. This command only gives the correct status of
        the LEDs if the last command was a 'night_light' command and not a 'rgb' light
        command, otherwise it gives the stored values of the 'night_light'.

        Example:
           {"is_on": false, "brightness": 0, "rgb": (0, 0, 0)}
        """
        state_int = self._gateway.send("get_night_light_rgb").pop()
        brightness = int_to_brightness(state_int)
        rgb = int_to_rgb(state_int)
        is_on = brightness > 0

        return {"is_on": is_on, "brightness": brightness, "rgb": rgb}

    @command(
        click.argument("brightness", type=int),
        click.argument("rgb", type=(int, int, int)),
    )
    def set_rgb(self, brightness: int, rgb: Tuple[int, int, int]):
        """Set gateway light using brightness and rgb tuple."""
        brightness_and_color = brightness_and_color_to_int(brightness, rgb)

        return self._gateway.send("set_rgb", [brightness_and_color])

    @command(
        click.argument("brightness", type=int),
        click.argument("rgb", type=(int, int, int)),
    )
    def set_night_light(self, brightness: int, rgb: Tuple[int, int, int]):
        """Set gateway night light using brightness and rgb tuple."""
        brightness_and_color = brightness_and_color_to_int(brightness, rgb)

        return self._gateway.send("set_night_light_rgb", [brightness_and_color])

    @command(click.argument("brightness", type=int))
    def set_rgb_brightness(self, brightness: int):
        """Set gateway light brightness (0-100)."""
        if 100 < brightness < 0:
            raise Exception("Brightness must be between 0 and 100")
        current_color = self.rgb_status()["rgb"]

        return self.set_rgb(brightness, current_color)

    @command(click.argument("brightness", type=int))
    def set_night_light_brightness(self, brightness: int):
        """Set night light brightness (0-100)."""
        if 100 < brightness < 0:
            raise Exception("Brightness must be between 0 and 100")
        current_color = self.night_light_status()["rgb"]

        return self.set_night_light(brightness, current_color)

    @command(click.argument("color_name", type=str))
    def set_rgb_color(self, color_name: str):
        """Set gateway light color using color name ('color_map' variable in the source
        holds the valid values)."""
        if color_name not in color_map.keys():
            raise Exception(
                "Cannot find {color} in {colors}".format(
                    color=color_name, colors=color_map.keys()
                )
            )
        current_brightness = self.rgb_status()["brightness"]

        return self.set_rgb(current_brightness, color_map[color_name])

    @command(click.argument("color_name", type=str))
    def set_night_light_color(self, color_name: str):
        """Set night light color using color name ('color_map' variable in the source
        holds the valid values)."""
        if color_name not in color_map.keys():
            raise Exception(
                "Cannot find {color} in {colors}".format(
                    color=color_name, colors=color_map.keys()
                )
            )
        current_brightness = self.night_light_status()["brightness"]

        return self.set_night_light(current_brightness, color_map[color_name])

    @command(
        click.argument("color_name", type=str),
        click.argument("brightness", type=int),
    )
    def set_rgb_using_name(self, color_name: str, brightness: int):
        """Set gateway light color (using color name, 'color_map' variable in the source
        holds the valid values) and brightness (0-100)."""
        if 100 < brightness < 0:
            raise Exception("Brightness must be between 0 and 100")
        if color_name not in color_map.keys():
            raise Exception(
                "Cannot find {color} in {colors}".format(
                    color=color_name, colors=color_map.keys()
                )
            )

        return self.set_rgb(brightness, color_map[color_name])

    @command(
        click.argument("color_name", type=str),
        click.argument("brightness", type=int),
    )
    def set_night_light_using_name(self, color_name: str, brightness: int):
        """Set night light color (using color name, 'color_map' variable in the source
        holds the valid values) and brightness (0-100)."""
        if 100 < brightness < 0:
            raise Exception("Brightness must be between 0 and 100")
        if color_name not in color_map.keys():
            raise Exception(
                "Cannot find {color} in {colors}".format(
                    color=color_name, colors=color_map.keys()
                )
            )

        return self.set_night_light(brightness, color_map[color_name])



import logging
from datetime import timedelta
from typing import Any, Dict, List

from .click_common import command, format_output
from .device import Device, DeviceStatus

_LOGGER = logging.getLogger(__name__)

ERROR_DESCRIPTION = [
    {
        "name": "Water temperature anomaly",
        "advice": "Check if inlet water temperature is among 5~38℃.",
    },
    {
        "name": "Inlet water flow meter damaged",
        "advice": "Try to purify water again after reinstalling the filter for serval times.",
    },
    {
        "name": "Water flow sensor anomaly",
        "advice": "Check if the water pressure is too low.",
    },
    {"name": "Filter life expired", "advice": "Replace filter."},
    {"name": "WiFi communication error", "advice": "Contact the after-sales."},
    {"name": "EEPROM communication error", "advice": "Contact the after-sales."},
    {"name": "RFID communication error", "advice": "Contact the after-sales."},
    {
        "name": "Faucet communication error",
        "advice": "Try to plug in the faucet again.",
    },
    {
        "name": "Purified water flow sensor anomaly",
        "advice": "Check whether all filters are properly installed and water pressure is normal.",
    },
    {
        "name": "Water leak",
        "advice": "Check if there is water leaking around the water purifier.",
    },
    {"name": "Floater anomaly", "advice": "Contact the after-sales."},
    {"name": "TDS anomaly", "advice": "Check if the RO filter is expired."},
    {
        "name": "Water temperature too high",
        "advice": "Check if inlet water is warm water with temperature above 40℃.",
    },
    {
        "name": "Recovery rate anomaly",
        "advice": "Check if the waste water pipe works abnormally and the RO filter is expired.",
    },
    {
        "name": "Outlet water quality anomaly",
        "advice": "Check if the waste water pipe works abnormally and the RO filter is expired.",
    },
    {
        "name": "Thermal protection for pumps",
        "advice": "The water purifier has worked for a long time, please use it after 20 minutes.",
    },
    {
        "name": "Dry burning protection",
        "advice": "Check if the inlet water pipe works abnormally.",
    },
    {
        "name": "Outlet water NTC anomaly",
        "advice": "Switch off the purifier and restart it again.",
    },
    {
        "name": "Dry burning NTC anomaly",
        "advice": "Switch off the purifier and restart it again.",
    },
    {
        "name": "Heater anomaly",
        "advice": "Switch off the purifier and restart it again.",
    },
]


class OperationStatus(DeviceStatus):
    def __init__(self, operation_status: int):
        """Operation status parser.

        Return value of operation_status: <int>

        We should convert the operation_status code to binary, each bit from
        LSB to MSB represents one error. It's able to cover multiple errors.

        Example operation_status value: 9 (binary: 1001)
        Thus, the purifier reports 2 errors, stands bit 0 and bit 3,
        means "Water temperature anomaly" and "Filter life expired".
        """
        self.err_list = [
            ERROR_DESCRIPTION[i]
            for i in range(0, len(ERROR_DESCRIPTION))
            if (1 << i) & operation_status
        ]

    @property
    def errors(self) -> List:
        return self.err_list


class WaterPurifierYunmiStatus(DeviceStatus):
    """Container for status reports from the water purifier (Yunmi model)."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Status of a Water Purifier C1 (yummi.waterpuri.lx11):

            [0, 7200, 8640, 520, 379, 7200, 17280, 2110, 4544,
            80, 4, 0, 31, 100, 7200, 8640, 1440, 3313]

        Parsed by WaterPurifierYunmi device as:
            {'run_status': 0, 'filter1_flow_total': 7200, 'filter1_life_total': 8640,
             'filter1_flow_used': 520, 'filter1_life_used': 379, 'filter2_flow_total': 7200,
             'filter2_life_total': 17280, 'filter2_flow_used': 2110, 'filter2_life_used': 4544,
             'tds_in': 80, 'tds_out': 4, 'rinse': 0, 'temperature': 31,
             'tds_warn_thd': 100, 'filter3_flow_total': 7200, 'filter3_life_total': 8640,
             'filter3_flow_used': 1440, 'filter3_life_used': 3313}
        """
        self.data = data

    @property
    def operation_status(self) -> OperationStatus:
        """Current operation status."""
        return OperationStatus(self.data["run_status"])

    @property
    def filter1_life_total(self) -> timedelta:
        """Filter1 total available time in hours."""
        return timedelta(hours=self.data["f1_totaltime"])

    @property
    def filter1_life_used(self) -> timedelta:
        """Filter1 used time in hours."""
        return timedelta(hours=self.data["f1_usedtime"])

    @property
    def filter1_life_remaining(self) -> timedelta:
        """Filter1 remaining time in hours."""
        return self.filter1_life_total - self.filter1_life_used

    @property
    def filter1_flow_total(self) -> int:
        """Filter1 total available flow in Metric Liter (L)."""
        return self.data["f1_totalflow"]

    @property
    def filter1_flow_used(self) -> int:
        """Filter1 used flow in Metric Liter (L)."""
        return self.data["f1_usedflow"]

    @property
    def filter1_flow_remaining(self) -> int:
        """Filter1 remaining flow in Metric Liter (L)."""
        return self.filter1_flow_total - self.filter1_flow_used

    @property
    def filter2_life_total(self) -> timedelta:
        """Filter2 total available time in hours."""
        return timedelta(hours=self.data["f2_totaltime"])

    @property
    def filter2_life_used(self) -> timedelta:
        """Filter2 used time in hours."""
        return timedelta(hours=self.data["f2_usedtime"])

    @property
    def filter2_life_remaining(self) -> timedelta:
        """Filter2 remaining time in hours."""
        return self.filter2_life_total - self.filter2_life_used

    @property
    def filter2_flow_total(self) -> int:
        """Filter2 total available flow in Metric Liter (L)."""
        return self.data["f2_totalflow"]

    @property
    def filter2_flow_used(self) -> int:
        """Filter2 used flow in Metric Liter (L)."""
        return self.data["f2_usedflow"]

    @property
    def filter2_flow_remaining(self) -> int:
        """Filter2 remaining flow in Metric Liter (L)."""
        return self.filter2_flow_total - self.filter2_flow_used

    @property
    def filter3_life_total(self) -> timedelta:
        """Filter3 total available time in hours."""
        return timedelta(hours=self.data["f3_totaltime"])

    @property
    def filter3_life_used(self) -> timedelta:
        """Filter3 used time in hours."""
        return timedelta(hours=self.data["f3_usedtime"])

    @property
    def filter3_life_remaining(self) -> timedelta:
        """Filter3 remaining time in hours."""
        return self.filter3_life_total - self.filter3_life_used

    @property
    def filter3_flow_total(self) -> int:
        """Filter3 total available flow in Metric Liter (L)."""
        return self.data["f3_totalflow"]

    @property
    def filter3_flow_used(self) -> int:
        """Filter3 used flow in Metric Liter (L)."""
        return self.data["f3_usedflow"]

    @property
    def filter3_flow_remaining(self) -> int:
        """Filter1 remaining flow in Metric Liter (L)."""
        return self.filter3_flow_total - self.filter3_flow_used

    @property
    def tds_in(self) -> int:
        """TDS value of input water."""
        return self.data["tds_in"]

    @property
    def tds_out(self) -> int:
        """TDS value of output water."""
        return self.data["tds_out"]

    @property
    def rinse(self) -> bool:
        """True if the device is rinsing."""
        return self.data["rinse"]

    @property
    def temperature(self) -> int:
        """Current water temperature in Celsius."""
        return self.data["temperature"]

    @property
    def tds_warn_thd(self) -> int:
        """TDS warning threshold."""
        return self.data["tds_warn_thd"]


class WaterPurifierYunmi(Device):
    """Main class representing the water purifier (Yunmi model)."""

    @command(
        default_output=format_output(
            "",
            "Operaton status: {result.operation_status}\n"
            "Filter1 total time: {result.filter1_life_total}\n"
            "Filter1 used time: {result.filter1_life_used}\n"
            "Filter1 remaining time: {result.filter1_life_remaining}\n"
            "Filter1 total flow: {result.filter1_flow_total} L\n"
            "Filter1 used flow: {result.filter1_flow_used} L\n"
            "Filter1 remaining flow: {result.filter1_flow_remaining} L\n"
            "Filter2 total time: {result.filter2_life_total}\n"
            "Filter2 used time: {result.filter2_life_used}\n"
            "Filter2 remaining time: {result.filter2_life_remaining}\n"
            "Filter2 total flow: {result.filter2_flow_total} L\n"
            "Filter2 used flow: {result.filter2_flow_used} L\n"
            "Filter2 remaining flow: {result.filter2_flow_remaining} L\n"
            "Filter3 total time: {result.filter3_life_total}\n"
            "Filter3 used time: {result.filter3_life_used}\n"
            "Filter3 remaining time: {result.filter3_life_remaining}\n"
            "Filter3 total flow: {result.filter3_flow_total} L\n"
            "Filter3 used flow: {result.filter3_flow_used} L\n"
            "Filter3 remaining flow: {result.filter3_flow_remaining} L\n"
            "TDS in: {result.tds_in}\n"
            "TDS out: {result.tds_out}\n"
            "Rinsing: {result.rinse}\n"
            "Temperature: {result.temperature} ℃\n"
            "TDS warning threshold: {result.tds_warn_thd}\n",
        )
    )
    def status(self) -> WaterPurifierYunmiStatus:
        """Retrieve properties."""

        properties = [
            "run_status",
            "f1_totalflow",
            "f1_totaltime",
            "f1_usedflow",
            "f1_usedtime",
            "f2_totalflow",
            "f2_totaltime",
            "f2_usedflow",
            "f2_usedtime",
            "tds_in",
            "tds_out",
            "rinse",
            "temperature",
            "tds_warn_thd",
            "f3_totalflow",
            "f3_totaltime",
            "f3_usedflow",
            "f3_usedtime",
        ]

        """
        Some models doesn't support a list of properties, while fetching them one
        per time usually runs into "ack timeout" error. Thus fetch them all at one
        time.
        Key "mode" (always 'purifying') and key "tds_out_avg" (always 0) are not
        included in return values.
        """
        values = self.send("get_prop", ["all"])

        prop_count = len(properties)
        val_count = len(values)
        if prop_count != val_count:
            _LOGGER.debug(
                "Count (%s) of requested properties does not match the "
                "count (%s) of received values.",
                prop_count,
                val_count,
            )

        return WaterPurifierYunmiStatus(dict(zip(properties, values)))
